/* *******************************************************************************************
The MIT License

Copyright (c) 2007 Steven G. Greenberg

Permission is hereby granted, free of charge, to any person obtaining a copy of this
software and associated documentation files (the "Software"), to deal in the Software
without restriction, including without limitation the rights to use, copy, modify, merge,
publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons
to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or
substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE
FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
******************************************************************************************* */

// ASF2VC1.cpp : Defines the entry point for the console application.
//

#include "stdafx.h"
#include <conio.h>

#include <wchar.h>

#include "log.h"

#include "constants.h"
#include "structs.h"
#include "FileCache2.h"
#include "AsfFile.h"
#include "AsfStream.h"
#include "Vc1Parse.h"

// local prototypes
BOOL			ParseCmndLine(int argc, wchar_t* argv[]);
BOOL			CmndLineFlagValOf(enum CMND_LINE_FLAG nFlagId);
HANDLE			CreateOutputFile(const wchar_t* pwszOutputFilename);
const _TCHAR*	WLastErrStr();
void			PrintHexByteSeq(struct VC1_HEADER_SEQUENCE* pSeq);

// ---------  cheesy globals --------------------------------------------------------------------------------
//
QWORD			g_qwTotalBytesWritten;	// was originally added for temporary diagnostics
DWORD			g_dwLogLevel;			// lazy man's method for this (value= 0, 1, 2 or 3). Will change this.
const wchar_t*	g_wpszInputFilename;	// lazy man's method for this, too
const wchar_t*	g_wpszOutputFilename;	// likewise

const char* pszUsage =
"Primary Usage: To Convert an ASF/WMV WVC1 input file to a VC-1 ES output file:\n\n"\
"  ASF2VC1 [<option(s)>] <input.wmv> <output.vc1>\n\n"\

"Other Usage:\n"\
"  ASF2VC1 <input.wmv>   Summary info only about WMV file <input.wmv>\n"\
"  ASF2VC1 <input.vc1>   Summary info only about VC-1 ES file <input.vc1>\n"\
"  ASF2VC1  --help       Display a list of other available <options>.\n";

const char* pszMoreHelp = "\n"\
"------- Output control flags. These affect the generated VC-1 file. ---------\n"\
"Default, and recommended: NONE of the below. See readme.txt for more info.\n\n"\

"-npar or --no-par          Ignore any PAR info found in the WMV file.\n"\
"-nfrm or --no-framerate    Ignore the framerate specified in the WMV file.\n"\
"-nseq or --no-repeat-sequence-header  Include VC-1 Sequence Hdr only once[1].\n\n"\

"---- Screen display level. Does not affect output file. Select only one. ----\n"\
"-v0 or --verbose=0  None except critical errors or prompts[2].\n"\
"-v1 or --verbose=1  Several lines of basic information.\n"\
"-v2 or --verbose=2  [default] All info incl compact \"char-per-frame\" table.\n"\
"-v3 or --verbose=3  Detailed \"line-per-frame\" listing replaces table above.\n\n"\

"Notes:\n"\

"[1] A concatenated VC-1 Sequence Header and Entry Point Hdr pair is inserted\n"\
"    prior to the first frame only; the remainder of the keyframes are preceded\n"\
"    by an Entry Point Hdr only. Default: Use the pair for before each keyframe.\n"\

"[2] -v0 messages (e.g. \"Overwrite File?\" prompts, etc.) are sent to stderr\n"\
"    and will appear on the console even if the app's command line output is\n"\
"    redirected (and hence will NOT appear in the redirected file). All other\n"\
"    output is sent to stdout to permit easy redirection to a text file.\n"\

"[3] \"nul:\" or \"nul\" may be specified as the output file to force a full demux\n"\
"    without actually producing any output. This is a convenient and fast way to\n"\
"    get extended details or simply to perform general testing.\n";

//   Maybe next version
// =======================
//--append-end-of-sequence
//--exclude-end-of-sequence
//--force-par=x
//--force-framerate=x
//--force-stream=x
//--first-frame=x
//--last-frame=x

int _tmain(int argc, wchar_t* argv[])
{
	HANDLE			hInput, hOutput;
	CAsfFile*		pAsfFileObject;
	DWORD			dwFramePeriod, dwParXVal, dwParYVal;
	CFileCache2*	pInputFileCache;
	struct VC1_HEADER_SEQUENCE	vc1OrigHdrs;
	struct VC1_HEADER_SEQUENCE	vc1ModHdrs;
	CVc1Parse		vc1Parser;
	BOOL			bSkipSeqStartCode, bCmndLineOk, bWantsHelp = FALSE;

	memset(&vc1OrigHdrs, 0, sizeof(vc1OrigHdrs));
	memset(&vc1ModHdrs, 0, sizeof(vc1ModHdrs));

	g_dwLogLevel = LOG_LEVEL_DEFAULT;

	Log(LOG_LEVEL_1, "\nASF2VC1 v1.2, build 20070526   (c)2007 Steven G Greenberg\n");
	Log(LOG_LEVEL_1, "Demuxes WMV WVC1 to VC-1 Elementary Stream. Advanced Profile Only.\n\n");

	bCmndLineOk = ParseCmndLine(argc, argv);
	bWantsHelp = CmndLineFlagValOf(CL_FLAG_HELP);

	if (!bCmndLineOk || (bCmndLineOk && bWantsHelp))
	{
		DWORD dwOutputType = bWantsHelp ? LOG_TYPE_CONSOLE_ERROR : LOG_LEVEL_1;

		Log(dwOutputType, pszUsage);

		if (bWantsHelp)		// Now, if he *asked* for help, we give him the whole nine yards....
			Log(dwOutputType, pszMoreHelp);
	}
	else	// -- else usage syntax is OK and it's not a help request
	{
		// Convert log flag into a global integer. This is a quick hack, and
		// should be changed to something a bit more sophisticated and flexible.
		//
		if (CmndLineFlagValOf(CL_FLAG_VERBOSE_3))
			g_dwLogLevel = 3;
		else if (CmndLineFlagValOf(CL_FLAG_VERBOSE_2))
			g_dwLogLevel = 2;
		else if (CmndLineFlagValOf(CL_FLAG_VERBOSE_1))
			g_dwLogLevel = 1;
		else if (CmndLineFlagValOf(CL_FLAG_VERBOSE_0))
			g_dwLogLevel = 0;


		// --------------------------------------------------------------------------------------
		hInput = CreateFileW(g_wpszInputFilename, GENERIC_READ, 0, NULL, OPEN_EXISTING, 0, NULL);
		if (hInput == INVALID_HANDLE_VALUE)
		{
			Log(LOG_TYPE_CONSOLE_ERROR, "Unable to open input file %S: %S\n", g_wpszInputFilename, WLastErrStr());
		}

		else	// ===========  opened an input file / create an associated ASF object============
		{
			// create an ASF object and associate it with the input file
			// in case we end up needing to demux, we'll need this input file cache object
			pInputFileCache = new CFileCache2;
			pInputFileCache->Init(hInput, 20);

			pAsfFileObject = new CAsfFile();
			pAsfFileObject->SetInputHandle(hInput);
			pAsfFileObject->SetCacheObject(pInputFileCache);


			if (!pAsfFileObject->ProcessHeader())
			{
				// ------------    Input file is NOT an ASF file -------------------
				Log(LOG_TYPE_CONSOLE_ERROR, "Note: Input file is not ASF, will try parsing it as VC-1 elementary...\n");
				Log(LOG_TYPE_CONSOLE_ERROR, "      (for information purposes only, as explained in \"Usage\".)\n");

				// Reset the file pointer for this informational attempt at parsing it as VC-1
				SetFilePointer(hInput, 0, NULL, FILE_BEGIN);

				ReadFile(hInput, vc1OrigHdrs.uchByteSeq, sizeof(vc1OrigHdrs.uchByteSeq),
							&(vc1OrigHdrs.dwLength), NULL);

				BOOL bIsVc1Elem = vc1Parser.BitStringToTable(&vc1OrigHdrs);
				if (bIsVc1Elem)
				{
					Log(LOG_TYPE_CONSOLE_ERROR, "File has VC-1 elementary stream headers, details follow:\n");
					vc1Parser.PrintBitParseTable();
				}
				else
				{
					Log(LOG_TYPE_CONSOLE_ERROR, "Not Recognized as VC-1 elementary, no Sequence / Entry Point headers Found\n");
				}
			}
			else
			{
				// ------------    Input file IS an ASF file ---------------------
				DWORD dwStreamIdxToWorkWith = pAsfFileObject->IndexOfMajorVidStream();

				if (dwStreamIdxToWorkWith > 0)	// if true, this guarantees it exists and the pointer below will be valid
				{
					// Get the appropriate stream object we're interested in from within the file object
					CAsfStream* pAsfStreamObject = pAsfFileObject->GetStreamObject(dwStreamIdxToWorkWith);

					pAsfStreamObject->GetAsfVc1Hdr(&vc1OrigHdrs);
					dwFramePeriod = pAsfStreamObject->GetAsfFramePeriod();

					Log(LOG_LEVEL_2, "\nSuccessfully retrieved %d bytes of VC-1 headers from video stream %d.\n",
						vc1OrigHdrs.dwLength, dwStreamIdxToWorkWith);
					Log(LOG_LEVEL_2, "Bytes as retrieved [concatenation of Sequence and Entry Point Headers]:\n");

					PrintHexByteSeq(&vc1OrigHdrs);

					// send bytes to the VC-1 bit parser
					vc1Parser.BitStringToTable(&vc1OrigHdrs);

					Log(LOG_LEVEL_2, "Interpretation of above VC-1 headers:\n");
					vc1Parser.PrintBitParseTable();

					// ------   info has been displayed -- are we supposed to demux this too? -------
					if (g_wpszOutputFilename != NULL && (hOutput = CreateOutputFile(g_wpszOutputFilename)) != INVALID_HANDLE_VALUE)
					{
						Log(LOG_LEVEL_1, "\nGenerating modified headers...\n");
						// We now "insert" frame rate as determined from the ASF file. "Legally",
						// this could be omitted from the VC-1 Elementary file, but in practice it makes
						// a file that's very "un-useful" and apparently rejected by various utilities
						// that ingest VC-1 files as their input. Reasonably enough, they can't work
						// with a file whose framerate is unknown.
						//
						if (CmndLineFlagValOf(CL_FLAG_NO_FRAMERATE))
							Log(LOG_LEVEL_1, "-- Note: Framerate info will be omitted [cmnd line request].\n");
						else
						{
							if (dwFramePeriod > 0)
							{
								Log(LOG_LEVEL_1, "-- Inserting Frame Period: %d uSecs (%2.3f FPS)...\n",
									dwFramePeriod, ((double)1000000.0/(double)dwFramePeriod));

								vc1Parser.SetFramePeriod(dwFramePeriod);
							}
							else
								Log(LOG_LEVEL_1, "-- Warning: No framerate found; it will be omitted from the output file\n");
						}
						// Now we insert PAR info if found. In contrast to above, it's fine if this
						// is not specified, presumably the player or utility which will ingest our
						// VC-1 file will simply assume 1:1 square pixels, which is probably the case.

						// Also add a non-square par if one was found in the ASF file. As mentioned above,
						// unlike the framerate insertion above, this is optional, and in fact is more the
						// exception than the rule. The info is usually absent from the ASF file, presumably
						// it's absence implies 1:1 "square" pixels. But we don't "assume" that, we just leave
						// it absent in the VC-1 if it's absent in the ASF, leaving the interpretation to others.
						//
						if (CmndLineFlagValOf(CL_FLAG_NO_PAR))
							Log(LOG_LEVEL_1, "-- Note: Any PAR info found will be omitted [cmnd line request].\n");
						else
						{
							pAsfStreamObject->GetAsfPar(&dwParXVal, &dwParYVal);

							if (dwParXVal == 0 || dwParYVal == 0)
								Log(LOG_LEVEL_1, "-- Note: PAR not specified in ASF input file- shall remain so in VC-1 output.\n");
							else
							{
								Log(LOG_LEVEL_1, "-- Inserting PAR of %d/%d (%1.3f)...\n", dwParXVal, dwParYVal,
									(double)dwParXVal / (double)dwParYVal);

								vc1Parser.SetPar(dwParXVal, dwParYVal);
							}
						}
						// Now generate and get our new freshly constructed Header Bytes sequence...
						//
						// Create a "full" (Sequence plus Entry Point) header sequence and supply it
						// to the ASF object to use during demux. This sequence MUST be used prior to
						// the first keyframe (the first frame) and may be used prior to every keyframe
						// (this is the default). Below we create a shorter sequence (Entry Point only)
						// which, as an option, can be substituted as the header prior to every keyframe
						// but the first. This latter option is legal, and was added mainly to test
						// create compatibility test files for comparison against some reference files
						// which are constructed that way, but is probably not preferable, despite the
						// fact that the  resulting file will be marginally smaller.

						vc1Parser.TableToBitString(&vc1ModHdrs, FALSE);		// Create a "full" (Seq & Entry point) hdr
						pAsfStreamObject->SetSeqAndEntryHdr(&vc1ModHdrs);	// ASF *will* demux will use this prior to first frame

						Log(LOG_LEVEL_2, "\nHeader bytes to be used in output [concatenated Seq and Entry Point Hdrs]:\n");
						PrintHexByteSeq(&vc1ModHdrs);

						vc1Parser.TableToBitString(&vc1ModHdrs, TRUE);		// Create a "short" (Entry point only) hdr
						pAsfStreamObject->SeqEntryHdrOnly(&vc1ModHdrs);		// User pref: ASF *may* use this prior to all other keyframes

						bSkipSeqStartCode = CmndLineFlagValOf(CL_FLAG_NO_REPEAT_SEQUENCE_HEADER);	// user's pref
						pAsfStreamObject->SetUseEntryHdrOnlyFlag(bSkipSeqStartCode);

						if (bSkipSeqStartCode)
						{
							Log(LOG_LEVEL_1, "-- Note: Seq Header will be included only once [cmnd line request].\n");
							Log(LOG_LEVEL_2, "--       Other keyframes will be preceded by Entry Point Hdr only:\n");
							PrintHexByteSeq(&vc1ModHdrs);
						}
						// ASF demux needs to know this to properly display frame types. This may seem confusing, this
						// info originally comes from the ASF file, and now we have to "feed it back". That's because
						// there's no native interlace support in ASF, so we figure it out from the original VC-1
						// headers that were there and then, after proper interpretation, we simply "tell it back"
						// (a lot more directly) what kind of a file it is. Hence:
						//
						pAsfStreamObject->SetInterlace(vc1Parser.IsInterlaced());	// tell it if it's interlaced

						if (TRUE)	// will be verbosity level
						{
							Log(LOG_LEVEL_2, "Interpretation of above VC-1 headers:\n");
							vc1Parser.PrintBitParseTable();
						}

						pAsfStreamObject->AsfDataDemux(hOutput);

						CloseHandle(hOutput);
						delete pInputFileCache;

					} // end "if output file was specified and created OK (and associated required input cache object created)

				} // a video stream was found

			} // end if "the input file was confirmed ASF"

			delete pAsfFileObject;
			CloseHandle(hInput);

		} // end "if we opened an input file and created associated ASF File Object"

	} // end "if usage syntax was OK"

	return (0);

} // end main

// -- command line options ---

struct OPTION Options[] =
{
	{CL_FLAG_HELP,						L"-h",		L"--help",						FALSE},
	{CL_FLAG_NO_PAR,					L"-npar",	L"--no-par",					FALSE},
	{CL_FLAG_NO_FRAMERATE,				L"-nfrm",	L"--no-framerate",				FALSE},
	{CL_FLAG_NO_REPEAT_SEQUENCE_HEADER,	L"-nseq",	L"--no-repeat-sequence-header",	FALSE},
	{CL_FLAG_VERBOSE_0,					L"-v0",		L"--verbose=0",					FALSE},
	{CL_FLAG_VERBOSE_1,					L"-v1",		L"--verbose=1",					FALSE},
	{CL_FLAG_VERBOSE_2,					L"-v2",		L"--verbose=2",					FALSE},
	{CL_FLAG_VERBOSE_3,					L"-v3",		L"--verbose=3",					FALSE},
	{CL_FLAG_END_OF_TABLE,				L"",		L"",							FALSE}
};


// ---------------------------------------------------------------------------------
// Read command line switches and filenames in any quantity in any order (well,
// the first filename is always "input" and the second, if any, is "output". And
// the filename cannot have "-" as its first character). Anyway, this isn't to fancy,
// the filenames are just set as globals for now and all the other flags are BOOLS.
// But they're all organized in a nice table with an ID to retrieve them, short and
// long form names, and specifiable defaults for each.
//
// The function returns currently returns FALSE if NO filenames are specified OR in
// the event any unknown options are encountered, else it returns TRUE. The status
// of the second (optional, output) filename should be checked by the caller checking
// if g_wpszOutputFilename is NULL or not.

BOOL ParseCmndLine(int argc, wchar_t* argv[])
{
	BOOL bRtnVal = TRUE, bRequestedHelp = FALSE;

	struct OPTION*	pOptPtr;

	for (int nArg = 1; nArg < argc; nArg++)
	{
		if (argv[nArg][0] != '-')	// not a "-"; assume a filename
		{
			if (g_wpszInputFilename == NULL)
				g_wpszInputFilename = argv[nArg];	// set first filename

			else if (g_wpszOutputFilename == NULL)
				g_wpszOutputFilename = argv[nArg];	// set second if first was already set
			else
				bRtnVal = FALSE;	// too many filenames
		}
		else
		{
			// The token DOES start with at least one "-"; search the switch table for matches (long or short form)
			for (pOptPtr = Options; pOptPtr->nFlag != CL_FLAG_END_OF_TABLE; pOptPtr++)
			{
				if ((_wcsicmp(argv[nArg], pOptPtr->wpszShortName) == 0) || (_wcsicmp(argv[nArg], pOptPtr->wpszLongName) == 0))
				{
					pOptPtr->bFlagVal = TRUE;
					break;
				}
			}
			// special cases
			if (pOptPtr->nFlag == CL_FLAG_HELP)		// set this local flag if the option is a help request
				bRequestedHelp = TRUE;
			else if (pOptPtr->nFlag == CL_FLAG_END_OF_TABLE)	// if loop ended, option is invalid
				bRtnVal = FALSE;

		} // end token is a valid switch search loop
	} // end loop cmnd line token ID loop

	// Return FALSE if any invalid options were specified or there wasn't at least one filename (unless the
	// user was *asking* for help, in which case no filename spec is normal)
	if (g_wpszInputFilename == NULL && !bRequestedHelp)
		bRtnVal = FALSE;

	return(bRtnVal);
}

// ------------------------------------------------------------------------------------------
// Return the flag val of any switches entered by the user. For generalization, the values
// are in a table and this does a "search", which probably takes all of a microsecond or
// something, but nonetheless good practice would be to retrieve these as needed but copy the
// value to a local variable for the one's that may get used repeatedly (e.g. every frame or whatever).
//
// For now they're all BOOL's, maybe next version we'll make this fancier

BOOL CmndLineFlagValOf(enum CMND_LINE_FLAG nFlagId)
{
	BOOL bRtnVal = FALSE;	// init val not used unless there's a "programmer error" (which ASSERTS at and of loop)
	struct OPTION*	pOptPtr;

	for (pOptPtr = Options; pOptPtr->nFlag != CL_FLAG_END_OF_TABLE; pOptPtr++)
	{
		if (pOptPtr->nFlag == nFlagId)
		{
			bRtnVal = pOptPtr->bFlagVal;
			break;
		}
	}
	_ASSERT_(pOptPtr->nFlag != CL_FLAG_END_OF_TABLE);
	return(bRtnVal);
}

// ========================================================================
void PrintHexByteSeq(struct VC1_HEADER_SEQUENCE* pSeq)
{
	for (DWORD dwIdx = 0; dwIdx < pSeq->dwLength; dwIdx++)
	{
		if (dwIdx != 0 && (dwIdx % 24) == 0)
			Log(LOG_LEVEL_2, "\n");

		Log(LOG_LEVEL_2, "%02x ", pSeq->uchByteSeq[dwIdx]);
	}
	Log(LOG_LEVEL_2, "\n\n");
}


HANDLE CreateOutputFile(const wchar_t* pwszOutputFilename)
{
	HANDLE	hOutput;
	DWORD	dwLastError;
	int nAnswer;

	hOutput = CreateFileW(pwszOutputFilename, GENERIC_WRITE, 0, NULL, CREATE_NEW, 0, NULL);

	if (hOutput == INVALID_HANDLE_VALUE)
	{
		dwLastError = GetLastError();
		if (dwLastError == ERROR_FILE_EXISTS)
		{
			Log(LOG_TYPE_CONSOLE_ERROR, "\nOutput file %S already exists - Overwrite? (y/N):", pwszOutputFilename);
			nAnswer = _getch();
			fputc(nAnswer, stderr);
			fputc('\n', stderr);

			if (toupper(nAnswer) == (int)'Y')
			{
				hOutput = CreateFileW(pwszOutputFilename, GENERIC_WRITE, 0, NULL, CREATE_ALWAYS, 0, NULL);

				if (hOutput == INVALID_HANDLE_VALUE)
				{
					Log(LOG_TYPE_CONSOLE_ERROR, "\nUnable to overwrite output file: %S", WLastErrStr());
				}
			}
		}
		else
		{
			Log(LOG_TYPE_CONSOLE_ERROR, "\nUnable to create output file: %S", WLastErrStr());
		}
	}
	return(hOutput);
}

const _TCHAR* WLastErrStr()
{
	static _TCHAR	wszErrorMsg[MAX_ERR_MSG_LEN];
	DWORD			dwLastError = GetLastError();

	FormatMessage(
			FORMAT_MESSAGE_FROM_SYSTEM,
			NULL,
			dwLastError,
			MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
			wszErrorMsg,
			MAX_ERR_MSG_LEN, NULL);

	return(wszErrorMsg);
}
