/* *******************************************************************************************
The MIT License

Copyright (c) 2007 Steven G. Greenberg

Permission is hereby granted, free of charge, to any person obtaining a copy of this
software and associated documentation files (the "Software"), to deal in the Software
without restriction, including without limitation the rights to use, copy, modify, merge,
publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons
to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or
substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE
FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
******************************************************************************************* */

#include "stdafx.h"

#include <malloc.h>
#include <objbase.h>
//#include <initguid.h>

#include "constants.h"
#include "structs.h"
#include "log.h"

#include "AsfGuids.h"
#include "AsfFile.h"
#include "AsfStream.h"


CAsfFile::CAsfFile()
{
	m_hInputFile = INVALID_HANDLE_VALUE;
	m_hOutputFile = INVALID_HANDLE_VALUE;

	memset(m_pStream, 0, sizeof(m_pStream));

	m_pInputFileCache = NULL;	// cache object

	m_dwPacketSize = 0;
}

CAsfFile::~CAsfFile(void)
{
	for (int nStream = 0; nStream < MAX_WMV_STREAMS; nStream++)
	{
		if (m_pStream[nStream] != NULL)
			delete m_pStream[nStream];
	}
}

CAsfStream* CAsfFile::GetStreamObject(DWORD dwStreamIdxToWorkWith)
{
	CAsfStream* pRtnVal = NULL;

	if (dwStreamIdxToWorkWith >= 1 && dwStreamIdxToWorkWith < MAX_WMV_STREAMS)
		pRtnVal = m_pStream[dwStreamIdxToWorkWith];

	return(pRtnVal);
}

// If successful, puts pair of headers in pVc1HdrBuf, allocated by caller
// dwMaxSizeofVc1HdrBuf is allocated size of that buffer
// Rtn val is length of actual data put into the buffer, or zero on failure
//
BOOL CAsfFile::ProcessHeader(void)
{
	struct ASF_MAIN_HDR_OBJ	asfMainHdr;
	struct ASF_ID_AND_SIZE	nextHdr;

	DWORD	dwBufIdx = 0;
	BYTE*	pBuf;
	DWORD	dwOurSizeLimit;
	BOOL	bReadRc;
	DWORD	dwRtnVal = 0;
	DWORD	dwBytesRead;

	SetFilePointer(m_hInputFile, 0, NULL, FILE_BEGIN);

	bReadRc = ReadFile(m_hInputFile, &asfMainHdr, sizeof(asfMainHdr), &dwBytesRead, NULL);

	if (bReadRc && dwBytesRead == sizeof(asfMainHdr) && asfMainHdr.asfIdAndSize.guidId == ASF_Header_Object)
	{
		Log(LOG_LEVEL_2, "Found ASF_Header_Object\n\n");
		dwOurSizeLimit = min((DWORD)asfMainHdr.asfIdAndSize.qwSize - sizeof(asfMainHdr), MAX_REASONABLE_HDR_SIZE);

		pBuf = (BYTE*)malloc(dwOurSizeLimit);

		if (pBuf != NULL)
		{
			// Read in everything else into a single buffer all at once
			bReadRc = ReadFile(m_hInputFile, pBuf, dwOurSizeLimit, &dwBytesRead, NULL);
			if (bReadRc && (dwBytesRead == dwOurSizeLimit))
			{
				// loop to skip by all the headers except the one we want, & stop when we find it
				for (DWORD dwHdrObj = 0; dwRtnVal == 0 && dwHdrObj < asfMainHdr.dwNumAddlHdrObjs &&
							dwBufIdx + sizeof(nextHdr) < dwOurSizeLimit; dwHdrObj++)
				{
					memcpy(&nextHdr, &pBuf[dwBufIdx], sizeof(nextHdr));

					if (dwBufIdx + (DWORD)nextHdr.qwSize <= dwOurSizeLimit)
					{
						// else it's safe to process, size wise - see if it's either of the two we're interested in
						if (nextHdr.guidId == ASF_Stream_Properties_Object)
						{
							dwRtnVal = ProcSomeStreamPropObj((struct STREAM_PROPERTIES_OBJECT*)(&pBuf[dwBufIdx]), FALSE);
						}
						else if (nextHdr.guidId == ASF_File_Properties_Object)
						{
							ProcFilePropObj((struct FILE_PROPERTIES_OBJECT*)(&pBuf[dwBufIdx]));
						}
						else if (nextHdr.guidId == ASF_Header_Extension_Object)
						{
							ProcASF_Header_Extension_Object((struct ASF_HEADER_EXTENSION_OBJECT *)(&pBuf[dwBufIdx]));
						}
					}
					dwBufIdx += (DWORD)nextHdr.qwSize;
				}
			}
			// the pointer should now point at the beginning of the data object

			free(pBuf);
		}
	}
	return (dwRtnVal);
}

DWORD CAsfFile::ProcFilePropObj(struct FILE_PROPERTIES_OBJECT* pFilePropObj)
{
	Log(LOG_LEVEL_2, "-- Found ASF_File_Properties_Object\n");
	// for now all we really need is this
	_ASSERT_(pFilePropObj->Minimum_Data_Packet_Size == pFilePropObj->Maximum_Data_Packet_Size);

	m_dwPacketSize = pFilePropObj->Maximum_Data_Packet_Size;

	return (0);
}

// Return the stream index of the video stream with the highest bitrate. Zero means no video
// streams were found, OK since zero is not a legal ASF stream number. In the unlikely event
// of a tie, the "last" video stream "wins"; makes the coding cleaner because be can use >=
// instead of > as the replace criteria so this function will return a video stream, if found,
// even if it has a bitrate of zero.
//
// call this function sometime AFTER ProcessHeader() or you won't get any results!
DWORD CAsfFile::IndexOfMajorVidStream()
{
	DWORD		dwRtnValStrmIdx = 0;
	DWORD		dwMaxBitrateFound = 0;
	CAsfStream* pStream;

	// search them al, there's nothing in the ASF spec, far as I know, that says stream numbers
	// have to be low and "contiguous" even though they always seem to be (e.g. typical example
	// has stream "1" is audio, stream "2" as one video stream and perhaps stream "3" as a secondary
	// video stream. But we'll search all 127 possible streams to be thorough.
	for (int dwStrmIdx = 1; dwStrmIdx < MAX_WMV_STREAMS; dwStrmIdx++)
	{
		pStream = m_pStream[dwStrmIdx];
		if (pStream != NULL && pStream->IsVideo())
		{
			if (pStream->GetAsfBitrate() >= dwMaxBitrateFound)	// highest one found?
			{
				dwMaxBitrateFound = pStream->GetAsfBitrate();
				dwRtnValStrmIdx = dwStrmIdx;		// yes, it's the new leader
			}
		}
	}
	if (dwRtnValStrmIdx == 0)
		Log(LOG_LEVEL_2, "No video streams found.\n");
	else
	{
		Log(LOG_LEVEL_2, "Video stream with highest bitrate is stream %d (bitrate = %d Kb/s)\n",
				dwRtnValStrmIdx, (dwMaxBitrateFound + 500)/1000);	// convert to Kb/s with roundoff for display
	}
	return(dwRtnValStrmIdx);
}


// -------------------------------------------------------------------------------------------------------------
// Also an aux routine called by GetHdrs() above, this one when an ASF_Header_Extension_Object is
// found. This one is fundamentally different from the one above, though. ProcStreamPropObj() gets
// called each time a ASF_Stream_Properties_Object is found, and processes it if its subtype matches,
// else it returns. The ProcASF_Header_Extension_Object here, by contrast, of which there is only one,
// that contains a number of "sub-objects" (which look like and in some case ARE the same as the higher
// level objects, but that's besides the point). So our job here is to search THRU this extension object
// and find the "sub-object" we're looking for, in our case a ASF_Extended_Stream_Properties_Object,
//
// Believe it or not, that seems to be the only way to get a answer to the seemingly basic question "what
// is the frame rate of this video?" (to be fair, I believe WMV supports per-frame "variable framerate",
// something we *should* not be running into. So the "average framerate", the value we want, I guess is
// almost an after thought from their perspective).
//
// Anyway, this searches thru the ASF_Header_Extension_Object for an ASF_Extended_Stream_Properties_Object.
//
// rev 1.1. update. Oh, and I seemed to missed the fact that, as convoluted as it sounds, the aforementioned
// ASF_Extended_Stream_Properties_Object may have a regular "outer level" ASF_Stream_Properties_Object at
// the end of it. Not only that, but it's probably the one we're looking for. Apparently they hide the new
// high bitrate streams from dumber old PC's or something.
//
BOOL CAsfFile::ProcASF_Header_Extension_Object(struct ASF_HEADER_EXTENSION_OBJECT * pExtHdrObj)
{
	BOOL				bRtnVal = FALSE;

	ASF_ID_AND_SIZE*	pNextObj;

	BYTE*				pPtr = (BYTE*)pExtHdrObj + sizeof(ASF_HEADER_EXTENSION_OBJECT);
	BYTE*				pPtrEnd = pPtr + pExtHdrObj->asfIdAndSize.qwSize;

	_ASSERT_(pExtHdrObj->guidReserved1 == ASF_Reserved_1);
	_ASSERT_(pExtHdrObj->wReserved2 == 6);

	Log(LOG_LEVEL_2, "-- Found ASF_Header_Extension_Object\n");

	// Believe it or not, the ASF documentation isn't clear about the size fields associated with this
	// unusual object. On one hand, it claims that the "Object Size" field is the size of this object
	// NOT including its contents, i.e. not including all the sub-objects. Sec. 3.1 "The value of this
	// field shall be set to 46 bytes (the size of the header itself).  On the other hand, the spec for
	// the "Header Extension Data Size" field says "It should be equal to the Object Size field minus
	// 46 bytes". Which means that field should always be zero if the first statement were true.

	// I think the first statement is wrong, and the Object Size field does include the included objects,
	// but we'll go with the Header Extension Data Size, as that's definitely not zero and does seem
	// to been created for the purpose we're using it for.

	_ASSERT_(pExtHdrObj->dwHeaderExtensionSize == (DWORD)pExtHdrObj->asfIdAndSize.qwSize - sizeof(ASF_HEADER_EXTENSION_OBJECT));

	while (pPtr + sizeof(ASF_ID_AND_SIZE) < pPtrEnd)
	{
		pNextObj = (ASF_ID_AND_SIZE*)pPtr;

		if (pPtr + pNextObj->qwSize < pPtrEnd)
		{
			if (pNextObj->guidId == ASF_Extended_Stream_Properties_Object)
			{
				bRtnVal = ProcSomeExtStreamPropObject((struct EXTENDED_STREAM_PROPERTIES_OBJECT*)pNextObj);
			}
			else if (pNextObj->guidId == ASF_Metadata_Object)
			{
				bRtnVal = ProcASF_Metadata_Object((struct METADATA_OBJECT*)pNextObj);
			}
		}
		pPtr += pNextObj->qwSize;
	}
	return (bRtnVal);
}


// ---------------------------------------------------------------------------------------------------
BOOL CAsfFile::ProcASF_Metadata_Object(struct METADATA_OBJECT* pMetadata)
{
	struct ASF_METADATA_DESCRIPTION_RECORD* pMetaDescripRec;

	WORD	wNumMetadataObjects = pMetadata->wRecordCount;
	DWORD	dwBytesLeft = (DWORD)pMetadata->asfIdAndSize.qwSize;

	BYTE*	pPtr = (BYTE*)pMetadata + sizeof(struct METADATA_OBJECT);
	dwBytesLeft -= sizeof(struct METADATA_OBJECT);

	//Log(LOG_LEVEL_2, "%d metadata objects...\n", wNumMetadataObjects);

	Log(LOG_LEVEL_2, "-- -- Found ASF_Metadata_Object; processing %d objects\n", wNumMetadataObjects);

	for (int nObj = 0; nObj < wNumMetadataObjects && dwBytesLeft > 0; nObj++)
	{
		if (dwBytesLeft < sizeof(ASF_METADATA_DESCRIPTION_RECORD))
		{
			_ASSERT_(0);
			break;
		}
		pMetaDescripRec = (struct ASF_METADATA_DESCRIPTION_RECORD*)pPtr;

		DWORD dwBytesUsed = ProcSomeMetaDataDescripRec(pMetaDescripRec, dwBytesLeft);
		pPtr += dwBytesUsed;
		dwBytesLeft -= dwBytesUsed;
	}
	//Log(LOG_LEVEL_2, "-- -- -- Note: bytes left after processing metadata: %d\n", dwBytesLeft);
	return (TRUE);
}

DWORD CAsfFile::ProcSomeMetaDataDescripRec(struct ASF_METADATA_DESCRIPTION_RECORD* pMetaDescripRec, DWORD dwBytesLeft)
{
	BOOL		dwBytesUsed = 0;
	CAsfStream* pStream = GetOrCreateAsfStream(pMetaDescripRec->wStreamNumber);

	if (pStream != NULL)
		dwBytesUsed = pStream->ProcMetaDataDescripRec(pMetaDescripRec, dwBytesLeft);

	return(dwBytesUsed);
}
// ===================================================================================================
// Data Demux
// ===================================================================================================

void CAsfFile::SetCacheObject(CFileCache2* pAsfFile)	// temp
{
	m_pInputFileCache = pAsfFile;
}


// this is a per-stream object, if we've never seen this stream before create it. Then call the routine to handle it.
BOOL CAsfFile::ProcSomeStreamPropObj(struct STREAM_PROPERTIES_OBJECT* pStreamPropObj, BOOL bExtraIndent)
{
	BOOL		bRtnVal = FALSE;
	CAsfStream*	pStream;

	pStream = GetOrCreateAsfStream(pStreamPropObj->wFlags & (MAX_WMV_STREAMS-1));

	if (pStream != NULL)
		bRtnVal = pStream->ProcStreamPropObj(pStreamPropObj, bExtraIndent);

	return(bRtnVal);
}

BOOL CAsfFile::ProcSomeExtStreamPropObject(struct EXTENDED_STREAM_PROPERTIES_OBJECT* pExtStreamObj)
{
	BOOL		bRtnVal = FALSE;
	CAsfStream*	pStream;

	pStream = GetOrCreateAsfStream(pExtStreamObj->dwStreamNumber);

	if (pStream != NULL)
		bRtnVal = pStream->ProcASF_Extended_Stream_Properties_Object(pExtStreamObj);

	return(bRtnVal);
}

CAsfStream* CAsfFile::GetOrCreateAsfStream(DWORD dwStreamNum)
{
	CAsfStream* pRtnVal = NULL;

	if (dwStreamNum == 0 || dwStreamNum > 127)
	{
		printf("Invalid stream #%d\n", dwStreamNum);
		assert(0);
	}
	else
	{
		if (m_pStream[dwStreamNum] == NULL)
			m_pStream[dwStreamNum] = new CAsfStream(m_pInputFileCache, dwStreamNum, m_dwPacketSize);	// create if needed

		pRtnVal = m_pStream[dwStreamNum];	// return it in either event
	}
	return(pRtnVal);	// existing stream, new stream, or NULL if dwStreamNum was out of range
}
