/* *******************************************************************************************
The MIT License

Copyright (c) 2007 Steven G. Greenberg

Permission is hereby granted, free of charge, to any person obtaining a copy of this
software and associated documentation files (the "Software"), to deal in the Software
without restriction, including without limitation the rights to use, copy, modify, merge,
publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons
to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or
substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE
FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
******************************************************************************************* */

#include "StdAfx.h"

#include <stdlib.h>

#include <objbase.h>
#include <initguid.h>

#include "constants.h"
#include "structs.h"
#include "log.h"

#include "AsfGuids.h"

#include "AsfStream.h"

extern QWORD	g_qwTotalBytesWritten;	// cheesy global
extern DWORD	g_dwLogLevel;			// cheesy global

//#define NO_OUTPUT

#ifdef NO_OUTPUT

BOOL _WriteFile(HANDLE hFile, LPCVOID lpBuffer, DWORD nNumberOfBytesToWrite, LPDWORD lpNumberOfBytesWritten, LPOVERLAPPED lpOverlapped)
{
	*lpNumberOfBytesWritten = nNumberOfBytesToWrite;
	g_qwTotalBytesWritten += nNumberOfBytesToWrite;
	return(TRUE);
}
#else
BOOL _WriteFile(HANDLE hFile, LPCVOID lpBuffer, DWORD nNumberOfBytesToWrite, LPDWORD lpNumberOfBytesWritten, LPOVERLAPPED lpOverlapped)
{
	BOOL bRc = WriteFile(hFile, lpBuffer, nNumberOfBytesToWrite, lpNumberOfBytesWritten, lpOverlapped);
	g_qwTotalBytesWritten += (QWORD)(*lpNumberOfBytesWritten);
	return(bRc);
}
#endif

struct GUID_TO_NAME_STRING	g_GuidNameTbl[] =
{
	{ &ASF_Audio_Media, "ASF_Audio_Media" },		// {F8699E40-5B4D-11CF-A8FD-00805F5C442B}
	{ &ASF_Video_Media,	"ASF_Video_Media" },	// {BC19EFC0-5B4D-11CF-A8FD-00805F5C442B}
	{ &ASF_Command_Media, "ASF_Command_Media" },	// {59DACFC0-59E6-11D0-A3AC-00A0C90348F6}
	{ &ASF_JFIF_Media, "ASF_JFIF_Media" },		// {B61BE100-5B4E-11CF-A8FD-00805F5C442B}
	{ &ASF_Degradable_JPEG_Media, "ASF_Degradable_JPEG_Media" }, // {35907DE0-E415-11CF-A917-00805F5C442B}
	{ &ASF_File_Transfer_Media, "ASF_File_Transfer_Media" },		// {91BD222C-F21C-497A-8B6D-5AA86BFC0185}
	{ &ASF_Binary_Media, "ASF_Binary_Media"},	// {3AFB65E2-47EF-40F2-AC2C-70A90D71D343}
	{ &CLSID_NULL, "" }							// {00000000-0000-0000-0000-000000000000}
};

CAsfStream::CAsfStream(CFileCache2* pOurInputFileCache, DWORD dwStreamNum, DWORD dwPacketSize)
{
	m_pInputFileCache = pOurInputFileCache;
	m_dwPacketSize = dwPacketSize;
	m_dwOurStreamNum = dwStreamNum;

	memset(&m_ExtStrProp, 0, sizeof(m_ExtStrProp));
	memset(m_PayloadExtSysArray, 0, sizeof(m_PayloadExtSysArray));

	memset(&m_AsfVc1Hdr, 0, sizeof(m_AsfVc1Hdr));
	memset(&m_Vc1SeqAndEntryHdr, 0, sizeof(m_Vc1SeqAndEntryHdr));
	memset(&m_Vc1EntryHdrOnly, 0, sizeof(m_Vc1EntryHdrOnly));

	// VC-1 related, when we do the demux. Prob should just make these local vars
	memset(m_dwFrameCounts, 0, sizeof(m_dwFrameCounts));

	m_dwMetaParWidth = 0;
	m_dwMetaParHeight = 0;

	m_bIsVideo = FALSE;
	m_bIsVc1 = FALSE;

	m_bIsInterlaced = FALSE;	// Should be set by the caller
	m_bUseEntryHdrOnly = FALSE;	// Should be set by the caller

	// VC-1 related, when we do the demux. Prob should just make these local vars
	m_dwTotalFrames = 0;
}

CAsfStream::~CAsfStream(void)
{
}

void CAsfStream::SeqEntryHdrOnly(const struct VC1_HEADER_SEQUENCE* pSeq)
{
	memcpy(&m_Vc1EntryHdrOnly, pSeq, sizeof(m_Vc1EntryHdrOnly));
}

void CAsfStream::SetSeqAndEntryHdr(const struct VC1_HEADER_SEQUENCE* pSeq)
{
	memcpy(&m_Vc1SeqAndEntryHdr, pSeq, sizeof(m_Vc1SeqAndEntryHdr));
}

void CAsfStream::GetAsfVc1Hdr(struct VC1_HEADER_SEQUENCE* pSeq)
{
	memcpy(pSeq, &m_AsfVc1Hdr, sizeof(m_AsfVc1Hdr));
}

void CAsfStream::GetAsfPar(DWORD* pdwHorizSize, DWORD* pdwVertSize)
{
	*pdwHorizSize = m_dwMetaParWidth;
	*pdwVertSize = m_dwMetaParWidth;
}

DWORD CAsfStream::GetAsfFramePeriod(void)
{
	return((DWORD) (m_ExtStrProp.qwAverageTimePerFrame +5)/10);
}

// -------------------------------------------------------------------------------------------
// Aux routine called by GetHdrs() above will actually analyze an ASF_Stream_Properties_Object
// when found. If it's the right type (an ASF_Video_Media object) and qualifies in other ways,
// the headers are written to pRsltBuf.
//
// Rtn val is zero on failure, or the number of bytes written if found.
// If the caller was too cheap to supply a big enough buffer, failure will result (it's only
// a couple of dozen bytes, for chrissakes!)

const char* CAsfStream::StreamNameOfGuid(GUID guid)
{
	const char* pszRtnVal = NULL;

	GUID_TO_NAME_STRING* pEntry;

	for (pEntry = g_GuidNameTbl; pEntry->pszName[0] != '\0' && pszRtnVal == NULL; pEntry++)
	{
		if (GUIDS_EQUAL(*(pEntry->pGuid), guid))
			pszRtnVal = pEntry->pszName;
	}
	// make this more convenient by never returning the word "<unknown" & never returning a NULL ptr
	if (pszRtnVal == NULL)
		pszRtnVal = "<unknown>";

	return(pszRtnVal);
}

BOOL CAsfStream::ProcStreamPropObj(struct STREAM_PROPERTIES_OBJECT* pStreamPropObj, BOOL bExtraIndent)
{
	DWORD	dwSeqAndEntryPointSize = 0;
	DWORD	dwStreamNum = (pStreamPropObj->wFlags) & (MAX_WMV_STREAMS-1);// 7 ls bits is stream number
	const char* pszBasePrefix = (bExtraIndent ? "-- -- -- " : "-- ");

	Log(LOG_LEVEL_2, "%sFound ASF_Stream_Properties_Object for ASF stream %d\n", pszBasePrefix, dwStreamNum);
	Log(LOG_LEVEL_2, "%sStream type: %s\n", pszBasePrefix, StreamNameOfGuid(pStreamPropObj->guidStream_Type));

	// copy the entire generic part of the stream properties object
	memcpy(&m_StreamProps, pStreamPropObj, sizeof(m_StreamProps));

	// see if its a video stream properties object, else forget it and just return
	if (pStreamPropObj->guidStream_Type == ASF_Video_Media)
	{
		m_bIsVideo = TRUE;
		BYTE* pEndOfGenericStreamObj = (BYTE*)pStreamPropObj + sizeof(struct STREAM_PROPERTIES_OBJECT);
		struct ASF_VIDEO_MEDIA_OBJECT*	pAsfVidMediaObj = (struct ASF_VIDEO_MEDIA_OBJECT*) pEndOfGenericStreamObj;

		char* psz4cc = (char*)(&pAsfVidMediaObj->biInfo.biCompression);
		Log(LOG_LEVEL_2, "%s-- FourCC: %c%c%c%c\n", pszBasePrefix, psz4cc[0], psz4cc[1], psz4cc[2], psz4cc[3]);

		if (pAsfVidMediaObj->biInfo.biCompression != DWLE_FOURCC("WVC1"))
		{
			Log(LOG_LEVEL_2, "%s-- Video stream has incorrect encoding format - should be \"WVC1\"\n", pszBasePrefix);
			Log(LOG_LEVEL_2, "%s-- Will not attempt to use this stream...\n", pszBasePrefix);
		}
		else
		{
			Log(LOG_LEVEL_2, "%s-- Stream correctly marked as VC-1 Advanced Profile\n", pszBasePrefix);
			Log(LOG_LEVEL_2, "%s-- Now looking for Sequence and Entry Point Bytes...\n", pszBasePrefix);
			m_bIsVc1 = TRUE;

			if ( pStreamPropObj->dwType_Specific_Data_Length > sizeof(struct ASF_VIDEO_MEDIA_OBJECT))
			{
				dwSeqAndEntryPointSize = pStreamPropObj->dwType_Specific_Data_Length - (sizeof(struct ASF_VIDEO_MEDIA_OBJECT));

				BYTE* pSeqDataStart = (BYTE*)(pAsfVidMediaObj) + sizeof(struct ASF_VIDEO_MEDIA_OBJECT);

				// finally! - copy the bytes.
				if (dwSeqAndEntryPointSize <= sizeof(m_AsfVc1Hdr.uchByteSeq))
				{
					Log(LOG_LEVEL_2, "%s-- Found Sequence and Entry Point Bytes; now copying...\n", pszBasePrefix);
					memcpy(m_AsfVc1Hdr.uchByteSeq, pSeqDataStart, dwSeqAndEntryPointSize);	// save data locally
					m_AsfVc1Hdr.dwLength = dwSeqAndEntryPointSize;
				}
			} // len is OK
		} // if format is correct
	}
	else
	{
		Log(LOG_LEVEL_2, "%s-- (Non-video stream; no further processing)\n", pszBasePrefix);
	}
	return(dwSeqAndEntryPointSize > 0);
}
// rev 1.1. update. As mentioned above, as convoluted as it sounds, this may have a regular
// ASF_Extended_Stream_Properties_Object at the end of it - processed the same, even though it's
// embedded several nested layers lower. It's probably the one we want, so we have to check for it.
//
//
// "Header Object"
// |
// +-- "Stream Properties object"  <======================= }
// |                                                        }
// +-- "Header Extension Object"                            }
// |     |                                                  }
// |     +-- "Extended Stream Properties Object"            } same object type, different location
// |     |    |                                             }
// |     |    + <various items>                             }
// |     |    |                                             }
// |     |    +-- "Stream Properties Object" <============= }

BOOL CAsfStream::ProcASF_Extended_Stream_Properties_Object(struct EXTENDED_STREAM_PROPERTIES_OBJECT* pExtStreamObj)
{
	if (pExtStreamObj->dwStreamNumber > 0 && pExtStreamObj->dwStreamNumber < MAX_WMV_STREAMS)
	{
		Log(LOG_LEVEL_2, "-- -- Found ASF_Extended_Stream_Properties_Object for ASF stream %d\n",
						pExtStreamObj->dwStreamNumber);

		if (pExtStreamObj->qwAverageTimePerFrame > 0)
		{
			DWORD dwAvgPeriodMicroSecs = (DWORD)((pExtStreamObj->qwAverageTimePerFrame + (QWORD)5) / (QWORD)10);

			Log(LOG_LEVEL_2, "-- -- -- avg object period: %d uSecs (%2.3f / sec)\n",
					dwAvgPeriodMicroSecs, ((double)1000000.0/(double)dwAvgPeriodMicroSecs));
		}
		else
			Log(LOG_LEVEL_2, "-- -- -- Warning: average object period is zero.");

		// other info of interest
		Log(LOG_LEVEL_2, "-- -- -- max object size: %d bytes\n", pExtStreamObj->dwMaximumObjectSize);
		Log(LOG_LEVEL_2, "-- -- -- leaky bucket size: %d bytes\n", pExtStreamObj->dwBufferSize);
		Log(LOG_LEVEL_2, "-- -- -- leaky bucket init fullness: %d bytes\n", pExtStreamObj->dwInitialBufferFullness);
		Log(LOG_LEVEL_2, "-- -- -- leaky bucket bitrate: %d KB/s\n", (pExtStreamObj->dwDataBitrate + 500)/1000);

		// copy to the stream array the entire fixed part of the structure, but not the variable length stuff
		memcpy(&m_ExtStrProp, pExtStreamObj, sizeof(struct EXTENDED_STREAM_PROPERTIES_OBJECT));

		// now it gets harder

		// use a 16-bit ptr for this part
		WORD* pwStreamNameInfo = (WORD*)((char*)pExtStreamObj + sizeof(struct EXTENDED_STREAM_PROPERTIES_OBJECT));

		// we're not saving this info, but we'll print it for diagnostics. We have to parse thru
		// it to get to the next thing to get to the next thing to finally get to the "hidden"
		// Stream Properties Object, which may or may not be there. This may be possible by using
		// some clever subtraction method or something, but this might as well take a look at this info.

		WORD wLangIdx, wStrmNameLen;
		wchar_t* pwszStrmName;

		for (int nStrmNameIdx = 0; nStrmNameIdx < pExtStreamObj->wStreamNameCount; nStrmNameIdx++)
		{
			wLangIdx = *pwStreamNameInfo++;
			wStrmNameLen = *pwStreamNameInfo++;
			pwszStrmName = (wchar_t*)pwStreamNameInfo;
			pwStreamNameInfo += (wStrmNameLen/sizeof(wchar_t));
			Log(LOG_LEVEL_2, "-- -- -- %d: Lang Idx %d, len %d, name: %S\r\n",
					nStrmNameIdx, wLangIdx, wStrmNameLen, pwszStrmName);
		}
		// ------------ now get Payload Extension Systems ---------------
		// switch back to a generic byte pointer, it's a mixture of data types coming up
		BYTE* puchPayload = (BYTE*)pwStreamNameInfo;

		for (DWORD dwPayExtIdx = 0; dwPayExtIdx < pExtStreamObj->wPayloadExtensionSystemCount; dwPayExtIdx++)
		{

			// copy GUID & advance pointer
			memcpy(&(m_PayloadExtSysArray[dwPayExtIdx].guid), puchPayload, sizeof(GUID));
			puchPayload += sizeof(GUID);

			// copy 16-bit data size value
			m_PayloadExtSysArray[dwPayExtIdx].wDataSize = *( (WORD*)puchPayload);
			puchPayload += sizeof(WORD);

			// copy 32-bit extension extra info length
			m_PayloadExtSysArray[dwPayExtIdx].dwExtInfoLen = *( (DWORD*)puchPayload);
			puchPayload += sizeof(DWORD);

			// now only get additional bytes if there are any, and limit the copy size
			DWORD dwBytesToCopy = max(m_PayloadExtSysArray[dwPayExtIdx].dwExtInfoLen, MAX_PAYLOAD_EXT_SYS_BYTES);

			// don't copy more than the max, but incr the full amount even if it's more than we copy
			memcpy(m_PayloadExtSysArray[dwPayExtIdx].uchExtInfo, puchPayload, dwBytesToCopy);
			puchPayload += m_PayloadExtSysArray[dwPayExtIdx].dwExtInfoLen;

			Log(LOG_LEVEL_2, "-- -- -- payload extension systems # %d:\n", dwPayExtIdx+1);
			Log(LOG_LEVEL_2, "-- -- -- -- Data size: %d   Bytes of extra info: %d\n",
					m_PayloadExtSysArray[dwPayExtIdx].wDataSize,
					m_PayloadExtSysArray[dwPayExtIdx].dwExtInfoLen);
		} // payload extension loop

		// with any luck, this MAY point to a ASF_Stream_Properties_Object,
		// {B7DC0791-A9B7-11CF-8EE6-00C00C205365}, if there are leftover bytes
		// check for hidden stream properties object

		//DWORD dwHiddenBytesLeft = (DWORD)(qwFinalIdx - qwExtraIdx);
		// I think that's all that can be stuck in here. Check if it is, and if it is, try processing it
		//ASSERT(dwHiddenBytesLeft == 0 || dwHiddenBytesLeft >= sizeof(struct STREAM_PROPERTIES_OBJECT) + (DWORD)HEADER_OVERHEAD);

		//if (dwHiddenBytesLeft >= sizeof(struct STREAM_PROPERTIES_OBJECT) + HEADER_OVERHEAD)
		//{
			// assume for the sec that it is, then immediately check for a GUID match. If it matches,
			// we the chances are less than 1 in 10^38 that it's a coincidence
			//GUID*	pGuidHiddenStrmProp = (GUID*)puchPayload;
			struct STREAM_PROPERTIES_OBJECT* pStreamPropObj = (struct STREAM_PROPERTIES_OBJECT*)puchPayload;

			if (GUIDS_EQUAL(pStreamPropObj->asfIdAndSize.guidId, ASF_Stream_Properties_Object))
			{
				Log(LOG_LEVEL_2, "-- -- -- Found \"Hidden\" Stream Properties Object, now processing...\n");
				ProcStreamPropObj(pStreamPropObj, TRUE);
			}
		//}
	}
	return(TRUE);
}

DWORD CAsfStream::ProcMetaDataDescripRec(struct ASF_METADATA_DESCRIPTION_RECORD* pMetaDescripRec, DWORD dwBytesLeft)
{
	BYTE*		pPtr = (BYTE*)pMetaDescripRec;
	void*		pVoidVal;
	char		szTempVal[META_TAG_DATA_MAX+1];
	WCHAR*		pwchName;

	if (pMetaDescripRec->wReserved == 0)
	{
		//Log(LOG_LEVEL_2, "Stream: %d, Namelen: %d, datatype:%d, datalen: %d\n",
			//pMetaDescripRec->wStreamNumber, pMetaDescripRec->wNameLength, pMetaDescripRec->wDataType, pMetaDescripRec->dwDataLength);

		pPtr += sizeof(ASF_METADATA_DESCRIPTION_RECORD);
		dwBytesLeft -= sizeof(ASF_METADATA_DESCRIPTION_RECORD);

		_ASSERT_(dwBytesLeft >= pMetaDescripRec->wNameLength);
		if (dwBytesLeft >= pMetaDescripRec->wNameLength)
		{
			pwchName = (WCHAR*)pPtr;

			pPtr += pMetaDescripRec->wNameLength;
			dwBytesLeft -= pMetaDescripRec->wNameLength;

			_ASSERT_(dwBytesLeft >= pMetaDescripRec->dwDataLength);
			if (dwBytesLeft >= pMetaDescripRec->dwDataLength)
			{
				pVoidVal = (void*) pPtr;

				// grab ones we're specifically interested in
				// this is coded for a single video stream and should really be changed,
				// we should have an array of them (like IsVBR below) and use the values
				// only when/if appropriate for the stream we're giving info for
				if (wcscmp(pwchName, L"AspectRatioX") == 0)
						m_dwMetaParWidth = *((DWORD*)pVoidVal);
				if (wcscmp(pwchName, L"AspectRatioY") == 0)
						m_dwMetaParHeight = *((DWORD*)pVoidVal);

				//wcstombs(m_pMiscInfo->metaTags[m_pMiscInfo->nNumMetaTags].szTagName, pwchName, META_TAG_NAME_MAX);
				//m_pMiscInfo->metaTags[m_pMiscInfo->nNumMetaTags].dwStreamNum = pMetaDescripRec->wStreamNumber;

				// now fill in the value for the meta-tag
				switch (pMetaDescripRec->wDataType)
				{
				  case 0:	// handle wide character string data
					//Log(LOG_LEVEL_2, "Value: %S\n", (WCHAR*) pVoidVal);
					wcstombs(szTempVal, (WCHAR*)pVoidVal, sizeof(szTempVal)-1);
					szTempVal[sizeof(szTempVal) -1] = '\0';
					break;

					// NOTE: the statement below is untrue, actually we do know the data length of any/all
					// of these - it's "pMetaDescripRec->dwDataLength". Maybe we'll do a better byte dump sometime
					// if we ever see these coming up.

				  case 1:	// handle unknown #of bytes of unknown data
					sprintf(szTempVal, "0x%02x, ...", (DWORD)(*((BYTE*)pVoidVal)));	// 1st byte of unknown #of bytes
					break;

				  case 2:	// a TRUE/FALSE value
					sprintf(szTempVal, "%s", (*((DWORD*)pVoidVal) == 0) ? "FALSE" : "TRUE");
					break;

				  case 3:	// a 32-bit numeric value
					sprintf(szTempVal, "%d", *((DWORD*)pVoidVal));
					break;

				  case 4:
					sprintf(szTempVal, "%I64d", *((QWORD *)pVoidVal));
					break;

				  case 5:	// a 16-bit numeric value
					sprintf(szTempVal, "%d", (DWORD)(*((WORD*)pVoidVal)));
					break;

				  default:
					strcpy(szTempVal, "*ERROR*");
					break;
				}
				Log(LOG_LEVEL_2, "-- -- -- Metadata for stream %d: %S = %s\n",
						pMetaDescripRec->wStreamNumber, pwchName, szTempVal);

				pPtr += pMetaDescripRec->dwDataLength;
				dwBytesLeft -= pMetaDescripRec->dwDataLength;
			}
		}
	} // reserved val is OK

	return((DWORD)(pPtr - (BYTE*)pMetaDescripRec));
}

#define DISP_FRAME_CHARS_PER_LINE	64


// they include this in the count, we subtract it since we're incrementing on the fly
#define HEADER_OVERHEAD	(sizeof(GUID) + sizeof (QWORD))

BOOL CAsfStream::AsfDataDemux(HANDLE hOutputFile)
{
	GUID	guidTopLevelDataObj;
	GUID	guidUniqueId;
	QWORD	qwTotalLenOfAllData;
	QWORD	qwNumDataPackets;
	DWORD	dwDataBytesLeft;
	BOOL	bRtnVal = FALSE;
	WORD	wReserved;

	m_pInputFileCache->GetNextGuid(&guidTopLevelDataObj);

	qwTotalLenOfAllData = m_pInputFileCache->NextQWordLe();

	dwDataBytesLeft = (DWORD)qwTotalLenOfAllData - (DWORD)HEADER_OVERHEAD;

	if (GUIDS_EQUAL(guidTopLevelDataObj, ASF_Data_Object))
	{
		// unique ID GUID
		m_pInputFileCache->GetNextGuid(&guidUniqueId);

		dwDataBytesLeft -= sizeof(GUID);

		// get total number of data objects
		qwNumDataPackets = m_pInputFileCache->NextQWordLe();

		dwDataBytesLeft -= sizeof(qwNumDataPackets);

		Log(LOG_LEVEL_2, "\nFound ASF_Data_Object\n");
		Log(LOG_LEVEL_2, "-- len = %d bytes; contains %d data packets\n",
					(DWORD)qwTotalLenOfAllData, (DWORD)qwNumDataPackets);
		// skip reserved
		wReserved = m_pInputFileCache->NextWordLe();

		dwDataBytesLeft -= sizeof(WORD);

		// first data object starts here
		Log(LOG_LEVEL_2, "Starting ASF demux / VC-1 write...\n\n");
		ProcessAsfDataObject(hOutputFile, dwDataBytesLeft);
	}
	return (bRtnVal);
}


void CAsfStream::ProcessAsfDataObject(HANDLE hOutputFile, DWORD dwDataLen)
{
	BYTE	uchEcFlag, uchPropFlags;
	DWORD	dwBytesLeft = dwDataLen;
	DWORD	dwNumPayloads, dwPayloadBytes, dwPayloadLenType;
	DWORD	dwFrameCountWriteIdx = 0;
	BYTE	uchLenFlags;
	QWORD	qwStartOfPacket;
	DWORD	dwBytesWritten;
	DWORD	dwDistFromLastKeyframe = 0, dwMaxDistFromLastKeyframe = 0;
	BOOL	bRc1, bRc2, bRc3;
	double	fFrameTime;

	enum VC1_FRAMETYPE	nFrameType;

	DWORD	dwBytesLeftInCurObj[MAX_WMV_STREAMS];
	memset(&dwBytesLeftInCurObj, 0, sizeof(dwBytesLeftInCurObj));

	// String for display (1 line) - DISP_FRAME_CHARS_PER_LINE plus terminating null
	char	chDispFrameChars[DISP_FRAME_CHARS_PER_LINE +1];
	memset(chDispFrameChars, 0, sizeof(chDispFrameChars));

	DWORD	dwFrameStartCode = NTOHL(FRM_SC);
	DWORD	dwEndOfSeqCode = NTOHL(EOS_SC);

	DWORD	dwMaxObjSizeForStream = m_ExtStrProp.dwMaximumObjectSize;

	memset(m_dwFrameCounts, 0, sizeof(m_dwFrameCounts));
	m_dwTotalFrames = 0;

	// This array must correspond 1:1 with the values of enum VC1_FRAMETYPE, as a var of that
	// type is cast to an integer which in turn is used to look up the display character.
	//
	// Assign an ASCII char to each frame type. This is for display purposes only, so it's not critical.
	char	chFrameTypeChar[FRMTYPE_NUM_FRMTYPES]
	= {

		'?', // FRMTYPE_UNDEFINED = 0,
		'|', // FRMTYPE_FRM_I,
		'-', // FRMTYPE_FRM_P,
		'<', // FRMTYPE_FRM_B,
		'^', // FRMTYPE_FRM_BI,
		'_', // FRMTYPE_FRM_D,
		'!', // FRMTYPE_FLD_I_I,
		'!', // FRMTYPE_FLD_I_P,
		'=', // FRMTYPE_FLD_P_I,
		'=', // FRMTYPE_FLD_P_P,
		'=', // FRMTYPE_FLD_B_B,
		'=', // FRMTYPE_FLD_B_BI,
		'=', // FRMTYPE_FLD_BI_B,
		'='  // FRMTYPE_FLD_BI_BI
	};

	char	chFrameTypeString[FRMTYPE_NUM_FRMTYPES][6]
	= {
		"?????", // FRMTYPE_UNDEFINED = 0,
		"I    ", // FRMTYPE_FRM_I,
		"P    ", // FRMTYPE_FRM_P,
		"B    ", // FRMTYPE_FRM_B,
		"BI   ", // FRMTYPE_FRM_BI,
		"SKIP ", // FRMTYPE_FRM_D,
		"I/I  ", // FRMTYPE_FLD_I_I,
		"I/P  ", // FRMTYPE_FLD_I_P,
		"P/I  ", // FRMTYPE_FLD_P_I,
		"P/P  ", // FRMTYPE_FLD_P_P,
		"B/B  ", // FRMTYPE_FLD_B_B,
		"B/BI ", // FRMTYPE_FLD_B_BI,
		"BI/B ", // FRMTYPE_FLD_BI_B,
		"BI/BI"  // FRMTYPE_FLD_BI_BI
	};

	// a way of expressing Table 35 in 7.1.1.4
	enum VC1_FRAMETYPE	frmBasedTypeCodes[16] =
	{
		FRMTYPE_FRM_P, FRMTYPE_FRM_P, FRMTYPE_FRM_P, FRMTYPE_FRM_P, FRMTYPE_FRM_P, FRMTYPE_FRM_P, FRMTYPE_FRM_P, FRMTYPE_FRM_P,	// '0xxx'
		FRMTYPE_FRM_B, FRMTYPE_FRM_B, FRMTYPE_FRM_B, FRMTYPE_FRM_B,	// '10xx'
		FRMTYPE_FRM_I, FRMTYPE_FRM_I,	// '110x'
		FRMTYPE_FRM_BI,		// '1110'
		FRMTYPE_FRM_D		// '1111'
	};
	// from table 105 in 9.1.1.42
	enum VC1_FRAMETYPE	fldBasedTypeCodes[8] =
	{
			FRMTYPE_FLD_I_I, FRMTYPE_FLD_I_P,  FRMTYPE_FLD_P_I,  FRMTYPE_FLD_P_P,
			FRMTYPE_FLD_B_B, FRMTYPE_FLD_B_BI, FRMTYPE_FLD_BI_B, FRMTYPE_FLD_BI_BI,
	};

	_ASSERT_ (IS_VALID_HANDLE(hOutputFile));

	_ASSERT_(dwMaxObjSizeForStream > 0 && dwMaxObjSizeForStream <= MAX_EXPECTED_FRAMESIZE);

	if (dwMaxObjSizeForStream == 0)
	{
		Log(LOG_LEVEL_2, "Warning: no max object size for stream %d; using default\n", m_dwOurStreamNum);
		dwMaxObjSizeForStream = MAX_EXPECTED_FRAMESIZE;
	}
	else if (dwMaxObjSizeForStream > MAX_EXPECTED_FRAMESIZE)
	{
		Log(LOG_LEVEL_2, "Warning: max object size for stream %d too large; using default\n", m_dwOurStreamNum);
		dwMaxObjSizeForStream = MAX_EXPECTED_FRAMESIZE;
	}

	BYTE*	pFrameBuf = (BYTE*)malloc(dwMaxObjSizeForStream);
	_ASSERT_(pFrameBuf != NULL);
	if (pFrameBuf == NULL)
		return;

	_ASSERT_(m_dwOurStreamNum > 0);
	if (m_dwOurStreamNum == 0)
		m_dwOurStreamNum = 1;	// shouldn't happen

	while (dwBytesLeft > 32)	// 32 is arbitrary for now
	{
		qwStartOfPacket = m_pInputFileCache->GetFilePos();

		uchEcFlag = m_pInputFileCache->NextByte();

		DWORD dwEcDataLen =			(DWORD) (uchEcFlag & 0x0f);
		DWORD dwOpaqueDataPresent =	(DWORD) (uchEcFlag & 0x10) >> 4;
		DWORD dwEcLenType =			(DWORD) (uchEcFlag & 0x60) >> 5;
		DWORD dwEcPresent1 =		(DWORD) (uchEcFlag & 0x80) >> 7;

		if (dwEcPresent1 == 1)
		{
			BYTE uchEcTypeAndNumber = m_pInputFileCache->NextByte();		// LS nybble is type
			BYTE uchEcCycle = m_pInputFileCache->NextByte();
			_ASSERT_(uchEcTypeAndNumber == 0 && uchEcCycle == 0);	// not strictly true, but we're not expecting ec

			uchLenFlags = m_pInputFileCache->NextByte();
		}
		else
			uchLenFlags = uchEcFlag;

		BOOL  bMultiplePayloads =	(uchLenFlags & 0x01) != 0;
		DWORD dwSeqLenType  =		(DWORD) (uchLenFlags & 0x06) >> 1;
		DWORD dwPadLenType  =		(DWORD) (uchLenFlags & 0x18) >> 3;
		DWORD dwPackLenType =		(DWORD) (uchLenFlags & 0x60) >> 5;
		DWORD dwEcPresent2 =		(DWORD) (uchLenFlags & 0x80) >> 7;

		_ASSERT_(dwEcPresent1 != dwEcPresent2);

		uchPropFlags = m_pInputFileCache->NextByte();

		DWORD dwRepDataLenType  =	(DWORD) (uchPropFlags & 0x03);
		DWORD dwOfstLenType  =		(DWORD) (uchPropFlags & 0x0c) >> 2;
		DWORD dwMediaObjLenType =	(DWORD) (uchPropFlags & 0x30) >> 4;
		DWORD dwStreamNumLenType =  (DWORD) (uchPropFlags & 0xc0) >> 6;

		DWORD dwPacketLen = m_pInputFileCache->NextAsfVarlen(dwPackLenType);
		DWORD dwSequenceLen = m_pInputFileCache->NextAsfVarlen(dwSeqLenType);
		DWORD dwPaddingLen = m_pInputFileCache->NextAsfVarlen(dwPadLenType);

		DWORD dwSendTime = m_pInputFileCache->NextDWordLe();
		WORD wDuration = m_pInputFileCache->NextWordLe();

		_ASSERT_(dwRepDataLenType == 1);

		if (bMultiplePayloads)
		{
			BYTE	uchPayloadFlags = m_pInputFileCache->NextByte();
			dwNumPayloads = (DWORD) (uchPayloadFlags & 0x3f);
			dwPayloadLenType = (DWORD) ((uchPayloadFlags & 0xc0) >> 6);
			_ASSERT_(dwPayloadLenType != 0);
		}
		else
		{
			dwNumPayloads = 1;
			dwPayloadLenType = 0;	// not used
		}

		for (DWORD dwPayLoad = 0; dwPayLoad < dwNumPayloads; dwPayLoad++)
		{
			DWORD dwStream = m_pInputFileCache->NextByte();
			BOOL bIsKeyFrame = (dwStream & 0x80) != 0;
			dwStream &= 0x7f;

			DWORD dwMediaObjectNum = m_pInputFileCache->NextAsfVarlen(dwMediaObjLenType);
			DWORD dwOffset = m_pInputFileCache->NextAsfVarlen(dwOfstLenType);
			DWORD dwReplicatedLen = m_pInputFileCache->NextAsfVarlen(dwRepDataLenType);

			if (dwReplicatedLen == 1)
			{
				Log(LOG_LEVEL_2, "Warning: ASF ReplicatedLen of '1' not handled yet; ignoring.\n");
			}
			else
			{
				_ASSERT_(dwReplicatedLen == 0 || dwReplicatedLen >= 8);
			}
			DWORD dwMediaObjSize = m_pInputFileCache->NextDWordLe();

			if (dwBytesLeftInCurObj[dwStream] == 0)
			{
				//Log(LOG_LEVEL_2, "New Media object has %d bytes in it\n", dwMediaObjSize);
				dwBytesLeftInCurObj[dwStream] = dwMediaObjSize;
			}
			_ASSERT_(dwOffset + dwBytesLeftInCurObj[dwStream] == dwMediaObjSize);

			DWORD dwPresentationTime = m_pInputFileCache->NextDWordLe();

			DWORD	dwExtraRepBytes = dwReplicatedLen - 2 * sizeof(DWORD);

			// adjust for extra junk in the replicated data array
			m_pInputFileCache->SkipBytes(dwExtraRepBytes);

			if (bMultiplePayloads)
				dwPayloadBytes = m_pInputFileCache->NextAsfVarlen(dwPayloadLenType);
			else
			{
				_ASSERT_(m_dwPacketSize != 0);
				QWORD qwCurFilePos = m_pInputFileCache->GetFilePos();

				dwPayloadBytes = m_dwPacketSize - (DWORD)(qwCurFilePos - qwStartOfPacket) - dwPaddingLen;
			}

			if (dwStream != m_dwOurStreamNum)
			{
				dwBytesLeftInCurObj[dwStream] -= dwPayloadBytes;
				m_pInputFileCache->SkipBytes(dwPayloadBytes);
			}
			else
			{
				_ASSERT_(dwFrameCountWriteIdx + dwPayloadBytes <= dwMaxObjSizeForStream);

				// ------- append bytes to buffer ------------
				if (dwFrameCountWriteIdx + dwPayloadBytes <= dwMaxObjSizeForStream)
				{
					m_pInputFileCache->ReadBytes(&pFrameBuf[dwFrameCountWriteIdx], dwPayloadBytes);
					dwFrameCountWriteIdx += dwPayloadBytes;
				}
				dwBytesLeftInCurObj[dwStream] -= dwPayloadBytes;

				if (dwBytesLeftInCurObj[dwStream] == 0)
				{
					// first take care of this little "GUI" issue. We display a char for each frame
					// processed, and each DISP_FRAME_CHARS_PER_LINE chars we display we do a CR/LF.
					if (m_dwTotalFrames != 0 && (m_dwTotalFrames % DISP_FRAME_CHARS_PER_LINE) == 0)
					{
						// update screen display with prev DISP_FRAME_CHARS_PER_LINE chars
						//
						// This is a bit of a hack, but we specifically have to inhibit the "graphical" form of
						// the display when the log level is at the surper-verbose (line-per-frame) list mode,
						// otherwise we'll have a mess. Everywhere else in the app (except once more a few lines
						// below) the Log() function takes care of itself just using the log level.
						if (g_dwLogLevel < 3)
							Log(LOG_LEVEL_2, "%s\n", chDispFrameChars);

						memset(chDispFrameChars, 0, sizeof(chDispFrameChars));	// re-clear for next line
					}

					// if the entire sequence is NOT interlaced, the frame type is in the first four bits.
					if (!m_bIsInterlaced)
						nFrameType = frmBasedTypeCodes[(pFrameBuf[0] & 0xf0) >> 4];
					else
					{
						// If the entire sequence is IS interlaced, the frame type might still be progressive,
						// or might be frame-based interlace, or might be field based interlaced.
						if ((pFrameBuf[0] & 0x80) == 0)		// progressive: acct for the extra bit & use table above
							nFrameType = frmBasedTypeCodes[(pFrameBuf[0] & 0x78) >> 3];

						// we already know bit 7 is set, only check bit 6 now
						else if((pFrameBuf[0] & 0x40) == 0)	// frame based interlace?
							nFrameType = frmBasedTypeCodes[(pFrameBuf[0] & 0x3c) >> 2];	// as above, but acct for 2 extra bits

						else	// else it's field based interlace, a whole separate animal.
						{
							// To obtain the [composite] frame type, we use a different table. The code is 3 bits
							// fixed length, and we obtain it after accounting for the two "1" bits that got us here.
							nFrameType = fldBasedTypeCodes[(pFrameBuf[0] & 0x38) >> 3];
						}
					}
					// Set a char representing frame type.
					//
					// Check for contradiction between WMV keyframe bit and VC1 frame type
					if (bIsKeyFrame && (nFrameType != FRMTYPE_FRM_I && nFrameType != FRMTYPE_FLD_I_I && nFrameType != FRMTYPE_FLD_I_P))
					{
						// use a special character in the display to indicate something unexpected happened
						// (ASSERT on debug builds)
						chDispFrameChars[m_dwTotalFrames % DISP_FRAME_CHARS_PER_LINE] = '*';
						_ASSERT_(0);
					}
					else
					{
						chDispFrameChars[m_dwTotalFrames % DISP_FRAME_CHARS_PER_LINE] = chFrameTypeChar[nFrameType];
					}

					// Verbose "per-frame" logging only
					//
					// The "signed" cast is to make VC6 happy. No matter, the frame time shouldn't exceed (2^63 / 10^7)
					// seconds -  I think that's about 29,000 years. Perhaps in some cosmic time lapse photography, though....
					//
					fFrameTime = ((double)((signed)m_ExtStrProp.qwAverageTimePerFrame) * (double)m_dwTotalFrames)/(double)1e7;

					// The following line of code, if the flag existed and was TRUE, would cause frame times displayed
					// on the long form (-v3) output screen to be offset by the "buffer size" - e.g. instead of starting
					// with 0.000 secs for the first frame it starts with, for example, 5.000 seconds.
					// Microsoft's internal converter displays it that way; setting this flag would make this app's display
					// compatible. Personally I think that's rather weird and unusual, though; I haven't seen that before, so
					// as of version 1.2 I didn't even include the flag or the command line interface to set it. I'm leaving
					// the commented line of code here, though, as a reminder & explanation of the difference.
					//
					//	if (m_bIncludePrebuf)
					//		fFrameTime += m_ExtStrProp.dwBufferSize/1000.0;		// commented out, see above
					//
					Log(LOG_LEVEL_3, "Frame:%6d  Time:%.3f sec  %c  Type:%s  Bytes:%d\n", m_dwTotalFrames+1,	// we'll use 1-based frame numbering
							fFrameTime,	bIsKeyFrame ? 'K' : ' ', chFrameTypeString[nFrameType], dwMediaObjSize);

					if (bIsKeyFrame)
					{
						// ----  is a keyframe ----
						dwDistFromLastKeyframe = 0;

						// Use m_Vc1SeqAndEntryHdr prior to the first keyframe, and, generally, prior to every
						// keyframe. However, if the caller set the m_bUseEntryHdrOnly flag, then all keyframes except the
						// very first ("m_dwTotalFrames == 0") will use the shorter "m_Vc1EntryHdrOnly"
						//
						if (m_dwTotalFrames == 0 || !m_bUseEntryHdrOnly)
							bRc1 = _WriteFile(hOutputFile, m_Vc1SeqAndEntryHdr.uchByteSeq, m_Vc1SeqAndEntryHdr.dwLength, &dwBytesWritten, NULL);
						else
							bRc1 = _WriteFile(hOutputFile, m_Vc1EntryHdrOnly.uchByteSeq, m_Vc1EntryHdrOnly.dwLength, &dwBytesWritten, NULL);

						_ASSERT_(bRc1 && dwBytesWritten == (m_dwTotalFrames == 0) ? m_Vc1SeqAndEntryHdr.dwLength : m_Vc1EntryHdrOnly.dwLength);
					}
					else
					{
						// ----  not a keyframe ----
						dwDistFromLastKeyframe++;

						if (dwDistFromLastKeyframe > dwMaxDistFromLastKeyframe)
							dwMaxDistFromLastKeyframe = dwDistFromLastKeyframe;

						// diagnostic message for the user. We'll continue anyway
						if (m_dwTotalFrames == 0)
							Log(LOG_LEVEL_1, "Error: First frame in file is not a keyframe\n");
					}
					// ---- all frames ----
					//
					// all frames get preceded by a "frame start code"
					bRc2 = _WriteFile(hOutputFile, &dwFrameStartCode, sizeof(dwFrameStartCode), &dwBytesWritten, NULL) ;
					_ASSERT_(bRc2 && dwBytesWritten == sizeof(dwFrameStartCode));

					bRc3 = _WriteFile(hOutputFile, pFrameBuf, dwMediaObjSize, &dwBytesWritten, NULL);
					_ASSERT_(bRc3 && dwBytesWritten == dwMediaObjSize);

					dwFrameCountWriteIdx = 0;
					m_dwFrameCounts[nFrameType]++;	// keep track of how many frames of each type
					m_dwTotalFrames++;
				}
			} // end is correct stream

			//m_pInputFileCache->SkipBytes(dwPayloadBytes);  << must be here if we are reading direct
		} // end multi-payload loop

		m_pInputFileCache->SkipBytes(dwPaddingLen);

		_ASSERT_(dwBytesLeft <= dwDataLen);
		dwBytesLeft = dwBytesLeft - (DWORD)(m_pInputFileCache->GetFilePos() - qwStartOfPacket);
		_ASSERT_(dwBytesLeft < dwDataLen);
	}
	// append an "end-of-sequence"
	_WriteFile(hOutputFile, &dwEndOfSeqCode, sizeof(dwEndOfSeqCode), &dwBytesWritten, NULL) ;

	// Write final line. See note above about this "if g_dwLogLevel < 3" hack. Will fix this later.
	if (g_dwLogLevel < 3)
		Log(LOG_LEVEL_2, "%s\n", chDispFrameChars);

	Log(LOG_LEVEL_1, "\nDone.\n\n");

	Log(LOG_LEVEL_2, "Summary of Frames Written:\n\n");

	// a few computations:

	DWORD	dwFieldKeyframes =		m_dwFrameCounts[FRMTYPE_FLD_I_I] + m_dwFrameCounts[FRMTYPE_FLD_I_P];

	DWORD	dwFieldNonKeyframes =	m_dwFrameCounts[FRMTYPE_FLD_P_I] + m_dwFrameCounts[FRMTYPE_FLD_P_P] +
									m_dwFrameCounts[FRMTYPE_FLD_B_B] + m_dwFrameCounts[FRMTYPE_FLD_B_BI] +
									m_dwFrameCounts[FRMTYPE_FLD_BI_B] + m_dwFrameCounts[FRMTYPE_FLD_BI_BI];

	DWORD	dwTotalKeyframes = m_dwFrameCounts[FRMTYPE_FRM_I] + dwFieldKeyframes;

	Log(LOG_LEVEL_2, "  Progressive or Frame Coded Interlaced Frames\n");
	Log(LOG_LEVEL_2, "  --------------------------------------------\n");

	if (g_dwLogLevel == 2)
	{
		Log(LOG_LEVEL_2, "  '|'  [I]  Intra-Coded (Keyframe)     : %d\n", m_dwFrameCounts[FRMTYPE_FRM_I]);
		Log(LOG_LEVEL_2, "  '-'  [P]  Predictive                 : %d\n", m_dwFrameCounts[FRMTYPE_FRM_P]);
		Log(LOG_LEVEL_2, "  '<'  [B]  Bidirectionally Predictive : %d\n", m_dwFrameCounts[FRMTYPE_FRM_B]);
		Log(LOG_LEVEL_2, "  '^'  [BI] Intra-Coded [B]-Like Frames: %d\n", m_dwFrameCounts[FRMTYPE_FRM_BI]);
		Log(LOG_LEVEL_2, "  '_'  [D]  Skipped (\"Duplicated\")     : %d\n", m_dwFrameCounts[FRMTYPE_FRM_D]);
	}
	else if (g_dwLogLevel == 3)
	{
		Log(LOG_LEVEL_3, "  [I]  Intra-Coded (Keyframe)       (K): %d\n", m_dwFrameCounts[FRMTYPE_FRM_I]);
		Log(LOG_LEVEL_3, "  [P]  Predictive                      : %d\n", m_dwFrameCounts[FRMTYPE_FRM_P]);
		Log(LOG_LEVEL_3, "  [B]  Bidirectionally Predictive      : %d\n", m_dwFrameCounts[FRMTYPE_FRM_B]);
		Log(LOG_LEVEL_3, "  [BI] Intra-Coded [B]-Like Frames     : %d\n", m_dwFrameCounts[FRMTYPE_FRM_BI]);
		Log(LOG_LEVEL_3, "  [D]  Skipped (\"Duplicated\")          : %d\n", m_dwFrameCounts[FRMTYPE_FRM_D]);
	}
	Log(LOG_LEVEL_2,     "\n  Field Coded Interlaced Frames\n");
	Log(LOG_LEVEL_2,     "  --------------------------------------------\n");

	if (g_dwLogLevel == 2)
	{
		Log(LOG_LEVEL_2, "  '!'  [I/I] or [I/P] (\"Keyframe\" Field Pairs): %d\n", dwFieldKeyframes);
		Log(LOG_LEVEL_2, "  '='  All VC-1 Supported \"Non-keyframe\" Pairs: %d\n", dwFieldNonKeyframes);
		Log(LOG_LEVEL_2, "\nNote: ';' includes [P/I], [P/P], [B/B], [B/BI], [BI/B], and [BI/BI].\n");
		Log(LOG_LEVEL_2, "      Use the \"-v3\" command line option for a specific breakdown.\n");
	}
	else if (g_dwLogLevel == 3)
	{
		Log(LOG_LEVEL_3, "  [I/I]   Field 1=[I],  Field 2=[I] (K): %d\n", m_dwFrameCounts[FRMTYPE_FLD_I_I]);
		Log(LOG_LEVEL_3, "  [I/P]   Field 1=[I],  Field 2=[P] (K): %d\n", m_dwFrameCounts[FRMTYPE_FLD_I_P]);
		Log(LOG_LEVEL_3, "  [P/I]   Field 1=[P],  Field 2=[I]    : %d\n", m_dwFrameCounts[FRMTYPE_FLD_P_I]);
		Log(LOG_LEVEL_3, "  [P/P]   Field 1=[P],  Field 2=[P]    : %d\n", m_dwFrameCounts[FRMTYPE_FLD_P_P]);
		Log(LOG_LEVEL_3, "  [B/B]   Field 1=[B],  Field 2=[B]    : %d\n", m_dwFrameCounts[FRMTYPE_FLD_B_B]);
		Log(LOG_LEVEL_3, "  [B/BI]  Field 1=[B],  Field 2=[BI]   : %d\n", m_dwFrameCounts[FRMTYPE_FLD_B_BI]);
		Log(LOG_LEVEL_3, "  [BI/B]  Field 1=[BI], Field 2=[B]    : %d\n", m_dwFrameCounts[FRMTYPE_FLD_BI_B]);
		Log(LOG_LEVEL_3, "  [BI/BI] Field 1=[BI], Field 2=[B]    : %d\n", m_dwFrameCounts[FRMTYPE_FLD_BI_BI]);

		Log(LOG_LEVEL_3, "  \nNote: Frame types designated (K) are considered keyframes.\n");
	}

	Log(LOG_LEVEL_1, "\nOverall summary:\n\n");

	Log(LOG_LEVEL_1,     "Interlaced Content     : %s\n", m_bIsInterlaced ? "Yes" : "No");
	Log(LOG_LEVEL_1,     "Max Keyframe Distance *: %d\n", dwMaxDistFromLastKeyframe +1);
	Log(LOG_LEVEL_1,     "Total Keyframes Written: %d\n", dwTotalKeyframes);
	Log(LOG_LEVEL_1,     "Total Frames Written   : %d\n", m_dwTotalFrames);
	Log(LOG_LEVEL_1,     "Total Bytes Written    : %d\n\n", (DWORD)g_qwTotalBytesWritten);	// displays wrong if >2^32; will fix later

	Log(LOG_LEVEL_1,     "* \"1-based\" - i.e. a file composed of 100%% keyframes displays \"1\", not \"0\".\n");

	free(pFrameBuf);
}

