/* *******************************************************************************************
The MIT License

Copyright (c) 2007 Steven G. Greenberg

Permission is hereby granted, free of charge, to any person obtaining a copy of this
software and associated documentation files (the "Software"), to deal in the Software
without restriction, including without limitation the rights to use, copy, modify, merge,
publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons
to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or
substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE
FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
******************************************************************************************* */

#include "stdafx.h"

#include "log.h"

#include "constants.h"
#include "structs.h"

#include "Vc1Parse.h"

struct VC1_PAR_VALUE
{
	const char*	pszDesig;
	DWORD		dwHoriz;
	DWORD		dwVert;
};


const struct VC1_PAR_VALUE g_Vc1ParTbl[] =
{
	{ "Unspecified",	0,  0  },		// 0
	{ "1:1(square)",	1,  1  },		// 1
	{ "12:11",			12, 11 },		// 2
	{ "10:11",			10, 11 },		// 3
	{ "16:11",			16, 11 },		// 4
	{ "40:33",			40, 33 },		// 5
	{ "24:11",			24, 11 },		// 6
	{ "20:11",			20, 11 },		// 7
	{ "32:11",			32, 11 },		// 8
	{ "80:33",			80, 33 },		// 9
	{ "18:11",			18, 11 },		// 10
	{ "15:11",			15, 11 },		// 11
	{ "64:33",			64, 33 },		// 12
	{ "160:99",			160,99 },		// 13
	{ "*ILLEGAL*",		0,  0  },		// 14
	{ "custom:",		0,  0  },		// 15
	{ "",				0,  0  }	// end of table
};

const char* g_pszXferCharTbl[] =
{ "* illegal *",
	"ITU-R BT.709-5, SMPTE 274M-2005, SMPTE 296M-2001, SMPTE 293M-2003",
	"not present", "* illegal *", "ITU-R BT. 1700 Part A [gamma = 2.2]",
	"ITU-R BT. 1700 Part B and Part C [gamma = 2.8]", "ITU-R BT.1700 Part A",
	"* illegal *", "ITU-R BT.1361 [Conventional Color Space]"
};

const struct PERIOD_TO_NR_DR g_Vc1NrDr[] =
{
	{41667,	1, 1 },    // 24.000 FPS
	{41708,	1, 2 },    // 23.976 FPS
	{40000,	2, 1 },    // 25.000 FPS
	{40040,	2, 2 },    // 24.975 FPS
	{33333,	3, 1 },    // 30.000 FPS
	{33367,	3, 2 },    // 29.970 FPS
	{20000,	4, 1 },    // 50.000 FPS
	{20020,	4, 2 },    // 49.950 FPS
	{16667,	5, 1 },    // 60.000 FPS
	{16683,	5, 2 },    // 59.940 FPS
	{20833,	6, 1 },    // 48.000 FPS
	{20854,	6, 2 },    // 47.952 FPS
	{13889,	7, 1 },    // 72.000 FPS
	{13903,	7, 2 },    // 71.928 FPS
	{0,		0, 0,}	   // end of table
};

const DWORD g_dwFrameRateNums[8] = {0, 24000, 25000, 30000, 50000, 60000, 48000, 72000 } ;

const DWORD g_dwFrameRateDenoms[3] = {0, 1000, 1001 };

CVc1Parse::CVc1Parse(void)
{
	m_dwFrmRateNR = 0;
	m_dwFrmRateDR = 0;
	m_wFrmRateEXP = 0;
	m_dwNumLeakyBucketVals = 0;

	memset(m_LeakyBucketVals, 0, sizeof(m_LeakyBucketVals));
	memset(m_HrdFull, 0, sizeof(m_HrdFull));

	// Initialize a non-const working table from const "initial" table. This fills it with
	// all of the various constants which represent the general structure, but leaves zeroes
	// where we will [perhaps] fill in values.
	memcpy(m_SeqAndEntryTbl, InitSeqAndEntryTbl, sizeof(InitSeqAndEntryTbl));
}

CVc1Parse::~CVc1Parse(void)
{
}

BOOL CVc1Parse::BitStringToTable(struct VC1_HEADER_SEQUENCE* pInputSeq)
{
	CGetBits InBits(pInputSeq->uchByteSeq, pInputSeq->dwLength);

	BOOL	bFoundSeqLayerHdr, bFoundEntryPointLayer = FALSE;
	DWORD	dwCompareVarValue;

	BOOL	bSkipEntries;
	int nIdx;

	bFoundSeqLayerHdr = InBits.ReSync32(SEQ_SC);
	nIdx = IndexOf(_SEQUENCE_START_) +1;

	while (m_SeqAndEntryTbl[nIdx].nParam != __EOT__)
	{
		switch (m_SeqAndEntryTbl[nIdx].nParam)
		{
		  case _LEAKY_BUCKETS_:
			FillLeakyBucketTable(&InBits);
			nIdx++;
			break;

		  case _END_OF_SEQ_HDR_:
			bFoundEntryPointLayer = InBits.ReSync32(ENT_SC);
			nIdx = IndexOf(_ENTRYPOINT_START_) +1;
			break;

		  case _HRD_FULLNESS_:
			FillHrdFullnessTable(&InBits);
			nIdx++;
			break;

		  default:
			// fill in value
			m_SeqAndEntryTbl[nIdx].dwValue = InBits.GetBits(m_SeqAndEntryTbl[nIdx].dwNumBits);
			m_SeqAndEntryTbl[nIdx].bValueIsValid = TRUE;

			// see if there's any conditional to be executed
			switch (m_SeqAndEntryTbl[nIdx].nCompareType)
			{
			  case BRA:	// branch always
				bSkipEntries = TRUE;
				break;
			  case BNE:	// branch if not equal

				if (m_SeqAndEntryTbl[nIdx].nParamToCompare == _____)	// means "this entry"
					dwCompareVarValue = m_SeqAndEntryTbl[nIdx].dwValue;
				else
					dwCompareVarValue = m_SeqAndEntryTbl[IndexOf(m_SeqAndEntryTbl[nIdx].nParamToCompare)].dwValue;

				bSkipEntries = dwCompareVarValue != m_SeqAndEntryTbl[nIdx].dwCompareTargVal;
				break;
			  default:
				bSkipEntries = FALSE;
			}
			if (bSkipEntries)
				nIdx = IndexOf(m_SeqAndEntryTbl[nIdx].nParamToSkipTo);
			else
				nIdx++;
		}
	}
	return (bFoundSeqLayerHdr && bFoundEntryPointLayer);
}

BOOL CVc1Parse::TableToBitString(struct VC1_HEADER_SEQUENCE* pOutputSeq, BOOL bSkipSeqStartCode)
{
	CPutBits OutBits(pOutputSeq->uchByteSeq, sizeof(pOutputSeq->uchByteSeq));

	DWORD	dwCompareVarValue;
	BOOL	bSkipEntries;
	int		nIdx;

	nIdx = IndexOf(bSkipSeqStartCode ? _ENTRYPOINT_START_ : _SEQUENCE_START_);

	while (m_SeqAndEntryTbl[nIdx].nParam != __EOT__)
	{
		switch (m_SeqAndEntryTbl[nIdx].nParam)
		{
		  case _LEAKY_BUCKETS_:
			DumpLeakyBucketTable(&OutBits);
			nIdx++;
			break;

		  case _END_OF_SEQ_HDR_:
			OutBits.PutBits(1, 0x1);
			OutBits.ResyncToNextByte();
			nIdx++;
			break;

		  case _HRD_FULLNESS_:
			DumpHrdFullnessTable(&OutBits);
			nIdx++;
			break;

		  default:
			// fill in value
			_ASSERT_(m_SeqAndEntryTbl[nIdx].bValueIsValid);

			OutBits.PutBits(m_SeqAndEntryTbl[nIdx].dwNumBits, m_SeqAndEntryTbl[nIdx].dwValue);

			// see if there's any conditional to be executed
			switch (m_SeqAndEntryTbl[nIdx].nCompareType)
			{
			  case BRA:	// branch always
				bSkipEntries = TRUE;
				break;
			  case BNE:	// branch if not equal

				  if (m_SeqAndEntryTbl[nIdx].nParamToCompare == _____)	// means "this entry"
					dwCompareVarValue = m_SeqAndEntryTbl[nIdx].dwValue;
				else
					dwCompareVarValue = m_SeqAndEntryTbl[IndexOf(m_SeqAndEntryTbl[nIdx].nParamToCompare)].dwValue;

				bSkipEntries = dwCompareVarValue != m_SeqAndEntryTbl[nIdx].dwCompareTargVal;
				break;
			  default:
				bSkipEntries = FALSE;
			}
			if (bSkipEntries)
				nIdx = IndexOf(m_SeqAndEntryTbl[nIdx].nParamToSkipTo);
			else
				nIdx++;
		}
	}
	OutBits.PutBits(1, 0x1);
	OutBits.ResyncToNextByte();
	pOutputSeq->dwLength = OutBits.GetByteLen();
	return (TRUE);
}


void CVc1Parse::FillLeakyBucketTable(CGetBits* pBitsIn)
{
	int nNumLeakyBuckets = m_SeqAndEntryTbl[IndexOf(HRD_NUM_LEAKY_BUCKETS)].dwValue;

	for (int nLeakyBucketIdx = 0; nLeakyBucketIdx < nNumLeakyBuckets; nLeakyBucketIdx++)
	{
		m_LeakyBucketVals[nLeakyBucketIdx].wHardRate = (WORD)pBitsIn->GetBits(16);
		m_LeakyBucketVals[nLeakyBucketIdx].wHardBuffer = (WORD)pBitsIn->GetBits(16);
	}
}

void CVc1Parse::DumpLeakyBucketTable(CPutBits* pBitsOut)
{
	int nNumLeakyBuckets = m_SeqAndEntryTbl[IndexOf(HRD_NUM_LEAKY_BUCKETS)].dwValue;

	for (int nLeakyBucketIdx = 0; nLeakyBucketIdx < nNumLeakyBuckets; nLeakyBucketIdx++)
	{
		pBitsOut->PutBits(16, m_LeakyBucketVals[nLeakyBucketIdx].wHardRate);
		pBitsOut->PutBits(16, m_LeakyBucketVals[nLeakyBucketIdx].wHardBuffer);
	}
}

void CVc1Parse::FillHrdFullnessTable(CGetBits* pBitsIn)
{
	int nNumLeakyBuckets = m_SeqAndEntryTbl[IndexOf(HRD_NUM_LEAKY_BUCKETS)].dwValue;

	for (int nLeakyBucketIdx = 0; nLeakyBucketIdx < nNumLeakyBuckets; nLeakyBucketIdx++)
		m_HrdFull[nLeakyBucketIdx] = (BYTE)pBitsIn->GetBits(8);
}

void CVc1Parse::DumpHrdFullnessTable(CPutBits* pBitsOut)
{
	int nNumLeakyBuckets = m_SeqAndEntryTbl[IndexOf(HRD_NUM_LEAKY_BUCKETS)].dwValue;

	for (int nLeakyBucketIdx = 0; nLeakyBucketIdx < nNumLeakyBuckets; nLeakyBucketIdx++)
		pBitsOut->PutBits(8, m_HrdFull[nLeakyBucketIdx]);
}

// ================================================================================================================
//                                      Misc general purpose table manipulation utilities
// ================================================================================================================


int CVc1Parse::IndexOf(enum VC1_HDR_PARAMS nTargParam)
{
	int nIdx = 0;

	while((m_SeqAndEntryTbl[nIdx].nParam != nTargParam) && (m_SeqAndEntryTbl[nIdx].nParam != __EOT__))
		nIdx++;

	_ASSERT_(m_SeqAndEntryTbl[nIdx].nParam != __EOT__);

	return(nIdx);
}

void CVc1Parse::SetTableValue(enum VC1_HDR_PARAMS nParm, DWORD dwVal)
{
	m_SeqAndEntryTbl[IndexOf(nParm)].dwValue = dwVal;
	m_SeqAndEntryTbl[IndexOf(nParm)].bValueIsValid = TRUE;
}

DWORD CVc1Parse::GetTableValue(enum VC1_HDR_PARAMS nParm)
{
	return(	m_SeqAndEntryTbl[IndexOf(nParm)].bValueIsValid ?
	m_SeqAndEntryTbl[IndexOf(nParm)].dwValue : 0);
}

// ================================================================================================================
//             Function(s) to "manually" get table entries with required by outside sources
// ================================================================================================================

BOOL CVc1Parse::IsInterlaced(void)
{
	return (GetTableValue(INTERLACE) != 0);
}

// ================================================================================================================
//             Functions to "manually" set table entries with data deduced from outside sources
// ================================================================================================================

void CVc1Parse::SetFramePeriod(DWORD dwMicroSecsPerFrame)
{
	// always return the FRAMERATEEXP method, no matter what. But if it's in the table, we'll set
	// NT and DR to non-zero values, and on return those will be used preferentially.

	double	fFrameRate = (double)MICROSECONDS_PER_SECOND / (double)dwMicroSecsPerFrame;
	double	fFrameRateExp = (32.0 * fFrameRate) -1.0;
	int		nFrameRateTblIdx;

	// Create an entire default display extension (only happens if this is the first function
	// called that sets a value in within the display extension, else simply returns doing =nothing)
	CreateDefaultDisplayExtension();

	// Regardless of which method we use below, we need to set the frame rate indicator flag
	SetTableValue(FRAMERATE_FLAG, 1);

	// check if it's in the table of "standard" (more or less!)  framerates
	for (nFrameRateTblIdx = 0; dwMicroSecsPerFrame != g_Vc1NrDr[nFrameRateTblIdx].dwPeriodMicroSecs &&
		g_Vc1NrDr[nFrameRateTblIdx].dwPeriodMicroSecs != 0; nFrameRateTblIdx++)
	{ }

	// if we *did* find it in the table set FRAMERATENR and FRAMERATEDR, and "set" the FRAMERATEIND flag to "0"
	if (g_Vc1NrDr[nFrameRateTblIdx].dwPeriodMicroSecs != 0)		// (we really only have to check either one of these)
	{
		SetTableValue(FRAMERATENR, g_Vc1NrDr[nFrameRateTblIdx].dwNumIdx);
		SetTableValue(FRAMERATEDR, g_Vc1NrDr[nFrameRateTblIdx].dwDenIdx);
		SetTableValue(FRAMERATEIND, 0);
	}
	else
	{
		// Otherwise don't set those - instead set FRAMERATEIND to "1" and set FRAMERATEEXP appropriately
		SetTableValue(FRAMERATEIND, 1);
		SetTableValue(FRAMERATEEXP, (DWORD) ((WORD)(fFrameRateExp +0.5)));	// the last .5 is for roundoff
	}
}

void CVc1Parse::SetPar(DWORD dwHoriz, DWORD dwVert)
{
	int		dwParTblIdx, nParTblMatchIdx = -1;

	_ASSERT_(dwHoriz > 0 && dwVert > 0);

	if ((dwHoriz > 0 && dwVert > 0))
	{
		// Create an entire default display extension (only happens if this is the first function
		// called that sets a value in within the display extension, else simply returns doing =nothing)
		CreateDefaultDisplayExtension();

		// Regardless of which method we use below, we need to set the frame rate indicator flag
		SetTableValue(ASPECT_RATIO_FLAG, 1);

		// check if it's in the table of "standard" (more or less!) of aspect ratios. The match
		// must be an exactly equal ratio, checked by cross-multiplication (high school algebra!)
		for (dwParTblIdx = 0; nParTblMatchIdx < 0
					&& g_Vc1ParTbl[dwParTblIdx].pszDesig[0] != '\0'; dwParTblIdx++)
		{
			// do the "cross multiply" check if this is a "real" take entry
			if (g_Vc1ParTbl[dwParTblIdx].dwVert > 0 && g_Vc1ParTbl[dwParTblIdx].dwHoriz > 0)
			{
				if (dwHoriz * g_Vc1ParTbl[dwParTblIdx].dwVert == dwVert * g_Vc1ParTbl[dwParTblIdx].dwHoriz)
					nParTblMatchIdx = dwParTblIdx;	// match found! will cause loop to break by itself
			}
		}
		// if we *did* find it in the table, just put the index of the match in ASPECT_RATIO.
		if (nParTblMatchIdx >= 0)
			SetTableValue(ASPECT_RATIO, (DWORD)nParTblMatchIdx);
		else
		{
			// an ASPECT_RATIO value of "15" means it's going to be specified explicitly
			SetTableValue(ASPECT_RATIO, 15);

			// Specify it explicitly. In the very unlikely event one or both of the ASF values
			// are greater than 256, we'll normalize our fraction setting the larger one to 255 and the
			// adjust the smaller one proportionally. In that result, that result is not guaranteed
			// to be the best integer approximation of the ratio we want, but this whole scenario is so
			// unlikely we're not going to worry about it (not to mention no one is going to care a PAR
			// error that's less than 0.4%)
			if (dwHoriz > 255 || dwVert > 255)
			{
				if (dwHoriz > dwVert)
				{
					dwVert = (255 * dwVert) / dwHoriz;
					dwHoriz = 255;
				}
				else
				{
					dwHoriz = (255 * dwHoriz) / dwVert;
					dwVert = 255;
				}
			}
			SetTableValue(ASPECT_HORIZ_SIZE, dwHoriz);
			SetTableValue(ASPECT_VERT_SIZE, dwVert);

		} // end if "specifying PAR explicitly"
	} // end "if supplied args were legal"
}

// This function should be called before any functions which mess around with values which are part of
// the Display Extension. This can (and should) be called prior to any function that does that, only the
// first call will fill in defaults. Once the defaults are in place, more specific functions can be called
// to install various values (there's no need to worry this function will "re-default" any such previously
// written values)

void CVc1Parse::CreateDefaultDisplayExtension(void)
{
	DWORD	dwHorizSize, dwVertSize;

	BOOL bDispExtAlreadyThere = (GetTableValue(DISPLAY_EXT) == 1);

	// If it's already been defaulted once, just don't touch it, as mentioned above.
	if (!bDispExtAlreadyThere)
	{
		// Else init it now with reasonable defaults
		SetTableValue(DISPLAY_EXT, 1);	// Set the flag itself, indicating its presence (and that it's been "defaulted");

		// default the display size to the coded size, making appropriate conversions for the different "inits" used
		dwHorizSize = 2 * GetTableValue(MAX_CODED_WIDTH)  +2;
		dwVertSize =  2 * GetTableValue(MAX_CODED_HEIGHT) +2;

		// ..and put those values in the table as defaults
		SetTableValue(DISP_HORIZ_SIZE, dwHorizSize -1);
		SetTableValue(DISP_VERT_SIZE, dwVertSize -1);

		// "set" frame rate flag to "not there". It may or may not be added later, in which case this flag's value will change
		SetTableValue(FRAMERATE_FLAG, 0);

		// as above for aspect ratio
		SetTableValue(ASPECT_RATIO_FLAG, 0);

		// as above for color format flag
		SetTableValue(COLOR_FORMAT_FLAG, 0);
	}
}

// ================================================================================================================
//                                                DEBUG / PRINT ROUTINES
// ================================================================================================================

void CVc1Parse::PrintBitParseTable(void)
{
	//Log(LOG_LEVEL_2, "\n---- Sequence Header ----\n");

	for (int nIdx = 0; m_SeqAndEntryTbl[nIdx].nParam != __EOT__; nIdx++)
	{

		if (m_SeqAndEntryTbl[nIdx].nParam == _END_OF_SEQ_HDR_)
		{
			PrintLeakyBucketTable();
		}
		else if (m_SeqAndEntryTbl[nIdx].dwNumBits > 0)
		{
			if (m_SeqAndEntryTbl[nIdx].bValueIsValid)
				PrintMiscEntry(&m_SeqAndEntryTbl[nIdx]);
			else
				Log(LOG_LEVEL_2, "%s: --\n", m_SeqAndEntryTbl[nIdx].pszParamName);
		}
	}
	PrintHrdFullnessTable();
}

void CVc1Parse::PrintMiscEntry(struct BIT_PARSE_TABLE* pEntry)
{
	DWORD	dwHorizSize, dwVertSize;
	DWORD	dwFrameRateNumIdx, dwFrameRateDenIdx;

	switch(pEntry->nParam)
	{
	  case LEVEL:
		if (pEntry->dwValue > 4)
			Log(LOG_LEVEL_2, "%s: %d (* illegal *)\n", pEntry->pszParamName, pEntry->dwValue);
		else
			Log(LOG_LEVEL_2, "%s: %d (AP3@L%1d)\n", pEntry->pszParamName, pEntry->dwValue, pEntry->dwValue);
		break;

	  case DISP_HORIZ_SIZE:
	  case DISP_VERT_SIZE:
		Log(LOG_LEVEL_2, "%s: %d (=%d)\n", pEntry->pszParamName, pEntry->dwValue, pEntry->dwValue + 1);
		break;

	  case MAX_CODED_WIDTH:
	  case MAX_CODED_HEIGHT:
	  case CODED_WIDTH:
	  case CODED_HEIGHT:
		Log(LOG_LEVEL_2, "%s: %d (=%d)\n", pEntry->pszParamName, pEntry->dwValue, 2 * pEntry->dwValue + 2);
		break;

	  case COLORDIFF_FORMAT:
		Log(LOG_LEVEL_2, "%s: %d (%s)\n", pEntry->pszParamName, pEntry->dwValue, (pEntry->dwValue == 1) ? "=4:2:0" : "* illegal *");
		break;

	  case ASPECT_RATIO:
		  Log(LOG_LEVEL_2, "%s: %d (PAR=%s)\n", pEntry->pszParamName, pEntry->dwValue, g_Vc1ParTbl[pEntry->dwValue].pszDesig);
		break;

	  case ASPECT_HORIZ_SIZE:
	  case ASPECT_VERT_SIZE:
		dwHorizSize = m_SeqAndEntryTbl[IndexOf(ASPECT_HORIZ_SIZE)].dwValue +1;
		dwVertSize = m_SeqAndEntryTbl[IndexOf(ASPECT_VERT_SIZE)].dwValue +1;
		Log(LOG_LEVEL_2, "%s: %d (PAR= %d/%d, =%1.3f)\n", pEntry->pszParamName, pEntry->dwValue,
			dwHorizSize, dwVertSize, (double)dwHorizSize/double(dwVertSize));
		break;

	  case FRAMERATENR:
	  case FRAMERATEDR:
		dwFrameRateNumIdx = m_SeqAndEntryTbl[IndexOf(FRAMERATENR)].dwValue;
		dwFrameRateDenIdx = m_SeqAndEntryTbl[IndexOf(FRAMERATEDR)].dwValue;
		if (dwFrameRateNumIdx == 0 || dwFrameRateNumIdx >= 8 ||
			dwFrameRateDenIdx == 0 || dwFrameRateDenIdx >= 3)
		{
			Log(LOG_LEVEL_2, "%s: %d (* illegal framerate *)\n", pEntry->pszParamName, pEntry->dwValue);
		}
		else
		{
			Log(LOG_LEVEL_2, "%s: %d (=%d/%d, =%2.3f FPS)\n", pEntry->pszParamName, pEntry->dwValue,
					g_dwFrameRateNums[dwFrameRateNumIdx], g_dwFrameRateDenoms[dwFrameRateDenIdx],
					(double)g_dwFrameRateNums[dwFrameRateNumIdx] / (double)g_dwFrameRateDenoms[dwFrameRateDenIdx]);
		}
		break;

	  case FRAMERATEEXP:
		Log(LOG_LEVEL_2, "%s: %d (= %2.3f FPS)\n", pEntry->pszParamName, pEntry->dwValue,
			(double)(m_SeqAndEntryTbl[IndexOf(FRAMERATEEXP)].dwValue +1)/(double)32.0);
		break;

	  case TRANSFER_CHAR:
		Log(LOG_LEVEL_2, "%s: %d (= %s)\n", pEntry->pszParamName, pEntry->dwValue,
			g_pszXferCharTbl[min(pEntry->dwValue, 8)]);
		break;

	  case _SEQUENCE_START_:
	  case _ENTRYPOINT_START_:
		Log(LOG_LEVEL_2, "%s", pEntry->pszParamName);	// Don't print the value of the start codes
		break;

	  default:
		Log(LOG_LEVEL_2, "%s: %d\n",  pEntry->pszParamName,  pEntry->dwValue);
		break;
	}
}
void CVc1Parse::PrintLeakyBucketTable(void)
{
	int nNumLeakyBuckets = m_SeqAndEntryTbl[IndexOf(HRD_NUM_LEAKY_BUCKETS)].dwValue;

	for (int nBucketIdx = 0; nBucketIdx < nNumLeakyBuckets; nBucketIdx++)
	{
		Log(LOG_LEVEL_2, "HRD_RATE[%d]: %d\n", nBucketIdx+1, (DWORD)m_LeakyBucketVals[nBucketIdx].wHardRate);
		Log(LOG_LEVEL_2, "HRD_BUFFER[%d]: %d\n", nBucketIdx+1, (DWORD)m_LeakyBucketVals[nBucketIdx].wHardBuffer);
	}
}

void CVc1Parse::PrintHrdFullnessTable(void)
{
	int nNumLeakyBuckets = m_SeqAndEntryTbl[IndexOf(HRD_NUM_LEAKY_BUCKETS)].dwValue;

	for (int nBucketIdx = 0; nBucketIdx < nNumLeakyBuckets; nBucketIdx++)
		Log(LOG_LEVEL_2, "HRD_FULL[%d]: %d\n", nBucketIdx+1, (DWORD)m_HrdFull[nBucketIdx]);
}

